"""Misc classes for the prevo package"""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>

import time
from threading import Thread
from random import random
from queue import Queue
from statistics import mean

import oclock
import numpy as np

# =========================== Dataname management ============================


def mode_to_names(mode, possible_names, default_names=None):
    """Determine active names as a function of input mode."""
    if mode is None:
        return [] if default_names is None else default_names
    names = []
    for name in possible_names:
        if name in mode:
            names.append(name)
    return names


# =========== Periodic Threaded systems for e.g. fake sensors etc. ===========


class PeriodicThreadedSystem:
    """Base class managing non-blocking, periodic control of devices."""

    name = None

    def __init__(self, interval=1, precise=False):
        """Parameters:

        - interval: update interval in seconds
        - precise (bool): use the precise option in oclock.Timer
        """
        self.timer = oclock.Timer(interval=interval, precise=precise)
        self.thread = None

    # ------------ Methods that need to be defined in subclasses -------------

    def _update(self):
        """Defined in subclass. Defines what needs to be done periodically."""
        pass

    def _on_start(self):
        """Defined in subclass (optional). Anything to do when system is started."""
        pass

    def _on_stop(self):
        """Defined in subclass (optional). Anything to do when system is stopped."""
        pass

    # ------------------------------------------------------------------------

    def _run(self):
        """Run _update() periodically, in a blocking fashion.

        See start() for non-blocking.
        """
        self._on_start()
        self.timer.reset()
        while not self.timer.is_stopped:
            self._update()
            self.timer.checkpt()
        self._on_stop()

    def start(self):
        """Non-blocking version of _run()."""
        self.thread = Thread(target=self._run)
        self.thread.start()

    def stop(self):
        self.timer.stop()
        self.thread.join()
        print(f'Non-blocking run of {self.name} stopped.')
        self.thread = None

    @property
    def dt(self):
        return self.timer.interval

    @dt.setter
    def dt(self, value):
        self.timer.interval = value


# ============ Classes to put sensor data in queues periodically =============


class PeriodicSensor(PeriodicThreadedSystem):
    """Read sensor periodically and put data in a queue with time info."""

    name = None         # Define in subclasses
    data_types = None   # Define in subclasses

    def __init__(self, interval=1):
        """Parameters:
        - interval: how often to read the sensor (in seconds)
        """
        super().__init__(interval=interval, precise=False)
        self.queue = Queue()

    def _read(self):
        """Define in subclasses. Must return data ready to put in queue."""
        pass

    def _update(self):
        data = self._read()
        self.queue.put(data)


class PeriodicTimedSensor(PeriodicSensor):
    """Automatically add information about time/duration of sensor reading."""

    def _read_sensor(self):
        """Define in subclass. Raw data from sensor, iterable if several channels."""
        pass

    def _read(self):
        with oclock.measure_time() as data:
            values = self._read_sensor()
        data['values'] = values
        data['name'] = self.name
        return data


# ======================== Dummy Sensors and Devices =========================


class DummyPressureSensor:
    """3 channels: 2 (random) pressures in Pa, 1 in mbar.

    Possibility of averaging when reading values.
    """

    def read(self, avg=1):
        val1, val2, val3 = [], [], []
        for _ in range(int(avg)):
            val1.append(3170 + random())
            val2.append(2338 + 2 * random())
            val3.append(17.06 + 0.5 * random())
        return {'P1 (Pa)': mean(val1),
                'P2 (Pa)': mean(val2),
                'P3 (mbar)': mean(val3)}


class DummyTemperatureSensor:
    """2 channels of (random) temperatures in °C.

    Possibility of averaging when reading values.
    """

    def read(self, avg=1):
        val1, val2 = [], []
        for _ in range(int(avg)):
            val1.append(25 + 0.5 * random())
            val2.append(22.3 + 0.3 * random())
        return {'T1 (°C)': mean(val1),
                'T2 (°C)': mean(val2)}


class DummyElectricalSensor:
    """Random electrical data returned as a numpy array with 3 columns
    corresponding to time and 2 electrical signals."""

    def __init__(self, interval=1, npts=100):
        self.interval = interval
        self.npts = npts

    def read(self):
        t0 = time.time() - self.interval
        time_array = t0 + np.linspace(start=0,
                                      stop=self.interval,
                                      num=self.npts)
        data_array_a = 0.1 * np.random.rand(self.npts) + 0.7
        data_array_b = 0.2 * np.random.rand(self.npts) + 0.3
        return np.vstack((time_array, data_array_a, data_array_b))


class DummyCirculatedBath:
    """Mimics a circulated bath to control temperature with a fluid."""

    def __init__(self, setpt=25):
        self.setpt = setpt
        self.status = 'off'

    def on(self):
        """Turn the bath on."""
        self.status = 'on'
        print('Bath ON')

    def off(self):
        """Turn the bath off."""
        self.status = 'off'
        print('Bath OFF')

    @property
    def tfluid(self):
        """Read fluid temperature in the bath."""
        return self.setpt + 0.1 * (random() - 0.5)


class DummyPump:

    def __init__(self):
        self._running = False
        self._normal_speed = False
        self._rpm = 0

    def on(self):
        self._running = True
        self._rpm = 30
        print('Pump on')

    def off(self):
        self._running = False
        self._rpm = 0
        print('Pump off')

    def speed(self, cmd):
        """Set the speed of the pump, can be 'full' or 'standby'."""
        if cmd == 'full':
            self._normal_speed = True
            self._rpm = 30
            print('Full speed')
        elif cmd == 'standby':
            self._normal_speed = False
            self._rpm = 7
            print('Standby speed')
        else:
            raise ValueError("Commands can only be 'full' or 'standby'")

    @property
    def status(self):

        return {'running': self._running,
                'rpm': self._rpm,
                'normal speed': self._normal_speed}



class DummyLapseCamera(PeriodicSensor):
    """Mock time-lapse camera returning white-noise images periodically"""

    name = 'Mock Lapse Camera'

    def _read(self):
        """Return image in a dict"""
        img = np.random.randint(256, size=(480, 640), dtype='uint8')
        return {'image': img}

    def read(self):
        """Return dict with image and timestamp"""
        with oclock.measure_time() as data:
            data['image'] = self._read()['image']
        return {'image': data['image'],
                'timestamp': data['time (unix)']}



"""Formatting sensor measurements giving numerical values.

Assumes data is saved with columns : time (unix), dt (s), m1, m2 ...
where m1, m2 are the different channels of the measurement.

Can be subclassed for arbitrary measurement formatting.
"""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>


# Standard library imports
from abc import ABC, abstractmethod

# Local imports
from .csv import CsvFile



# ======================= Base classes for saved data ========================
# ------------------- (used for plotting, loading, etc.) ---------------------


class SavedDataBase(ABC):
    """Abstract base class for live measurements of sensors"""

    def __init__(self, name):
        """Parameters:
        - name: name of sensor/recording
        """
        self.name = name
        self.data = None

    @abstractmethod
    def load(self, nrange=None):
        """Load measurement from saved data (time, etc.) into self.data

        nrange = None should load all the data
        nrange = (n1, n2) loads measurement numbers from n1 to n2 (both
        included), and first measurement is n=1.
        """
        pass

    @abstractmethod
    def number_of_measurements(self):
        """Return total number of measurements currently saved in file."""
        pass

    @abstractmethod
    def format_as_measurement(self):
        """Transform loaded data into something usable (e.g. by plots etc.)"""
        pass


# ========================== Examples of subclasses ==========================


class SavedCsvData(CsvFile, SavedDataBase):
    """Class managing saved measurements to CSV files (with pandas)"""

    def __init__(self, name, filename, path='.', csv_separator='\t'):

        SavedDataBase.__init__(self, name=name)

        CsvFile.__init__(self, filename=filename, path=path,
                         csv_separator=csv_separator)

    def load(self, nrange=None):
        self.data = CsvFile.load(self, nrange=nrange)

    def format_as_measurement(self):
        """Generate useful attributes for plotting on a Graph() object.

        Here we assume that the first two columns in the csv represent
        - time (unix)
        - dt (s)
        and are considered as time columns, not value columns
        """
        measurement = {}
        measurement['name'] = self.name
        measurement['time (unix)'] = self.data['time (unix)'].values
        # remove time columns
        measurement['values'] = [column.values for _, column in self.data.iloc[:, 2:].items()]
        return measurement



"""Mock camera device for tests."""


# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>


from abc import ABC, abstractmethod
import tkinter as tk
import time
from queue import Queue, Empty
from threading import Thread, Event
import itertools
from traceback import print_exc

import numpy as np
from PIL import Image, ImageTk

try:
    import cv2
except ModuleNotFoundError:
    pass

import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation

# ====================== Parameters for Tkinter viewers ======================


bgcolor = '#485a6a'
textcolor = '#e7eff6'
fontfamily = "serif"


# =============================== Base classes ===============================


class ViewerBase(ABC):

    def __init__(self, dt_graph):
        self.dt_graph = dt_graph

    @abstractmethod
    def run(self):
        """Indicate how to run the live viewer."""
        pass

    def _init_window(self):
        """How to create window."""
        pass

    def _init_run(self):
        """Anything to be done just before starting the viewer."""
        pass

    def start(self):
        try:
            self._init_window()
            self._init_run()
            self.run()
        except Exception:
            print('--- !!! Error in Viewer !!! ---')
            print_exc()
        self.on_stop()

    def on_stop(self):
        """Things to do when viewer is stopped/closed"""
        # This is e.g. to stop camera readings in live() when window is closed.
        # In record() situations, this needs to be subclassed to avoid stopping
        # the recording when closing the window.
        self.e_stop.set()


class SingleViewer(ViewerBase):
    """Base class for GUIs for viewing images from image queues."""

    def __init__(self,
                 image_queue,
                 e_stop=None,
                 dt_graph=0.01,
                 info_queue=None,
                 name='Camera'):
        """Parameters:

        - image_queue: queue in which taken images are put.
        - e_stop: stopping event (threading.Event or equivalent)
        - dt_graph: how often (in seconds) the viewer is updated
        - info_queue: if not None, print info (received as str) below images
        - name: optional name for display purposes.
        """
        super().__init__(dt_graph=dt_graph)
        self.image_queue = image_queue
        self.info_queue = info_queue
        self.e_stop = e_stop if e_stop is not None else Event()
        self.name = name

        # store times at which images are shown on screen (e.g. for fps calc.)
        self.display_times = []             # to calculate fps on all times
        self.display_times_queue = Queue()  # to calculate fps on partial data

    def _measurement_to_image(self, measurement):
        """How to transform individual elements from the queue into an image.

        (returns an array-like object).
        Can be subclassed to accommodate different queue formats.
        """
        return measurement['image']

    def _store_display_times(self):
        t = time.perf_counter()
        self.display_times.append(t)
        self.display_times_queue.put(t)

    def _display_info(self, info, *args, **kwargs):
        """How to display information from info_queue on image.

        Define in subclasses.
        """
        pass

    def _manage_info(self, *args, **kwargs):
        """Get information from info_queue and display it if not None."""
        if self.info_queue:
            info = get_last_from_queue(self.info_queue)
            if info:
                self._display_info(info, *args, **kwargs)


class SingleStreamViewer:
    """Add functionality to SingleViewer for showing live camera streams."""

    def __init__(self, *args, display_fps=True, dt_check=2, **kwargs):
        """Parameters:

        - display_fps: if True, indicate current display fps on viewer
        - dt_check: how often (in seconds) display fps are calculated
        """
        info_queue = Queue() if display_fps else None
        self.dt_check = dt_check
        super().__init__(*args, **kwargs, info_queue=info_queue)

    def _init_run(self):
        """If things need to be done before running in subclasses"""
        super()._init_run()
        if self.info_queue:
            fps_calculator = LiveFpsCalculator(time_queue=self.display_times_queue,
                                               info_queue=self.info_queue,
                                               e_stop=self.e_stop,
                                               dt_check=self.dt_check)
            fps_calculator.start()

    def on_stop(self):
        """What to do when live view is stopped"""
        super().on_stop()
        if self.display_times:
            fps = LiveFpsCalculator._calculate_fps(self.display_times)
            print(f'Average display frame rate [{self.name}]: {fps:.3f} fps')


class MultipleViewer(ViewerBase):
    """Display several image streams at the same time."""

    def __init__(self,
                 image_queues,
                 e_stop=None,
                 dt_graph=0.01,
                 Viewer=SingleViewer,
                 add_ppties=None,
                 add_ppty_name=None,
                 **kwargs):
        """Parameters:

        - image_queues: dict {camera name: queue in which taken images are put.}
        - e_stop: stopping event (threading.Event or equivalent)
        - dt_graph: how often (in seconds) the GUI window is updated
        - Viewer: which SingleViewer to use
        - add_ppties: any additional properties as dict {name: ppty} to pass
          to the viewer, for properties that depend on name for Viewer init.
        - add_ppty_name: str: name of pppty (key in kwargs for viewer)
        - **kwargs: any other optional keyword arguments required by the Viewer.
        """
        super().__init__(dt_graph=dt_graph)

        self.image_queues = image_queues
        self.e_stop = e_stop if e_stop is not None else Event()

        self.viewers = {}

        for name, image_queue in self.image_queues.items():

            if add_ppties is not None:
                add_ppty = {add_ppty_name: add_ppties[name]}
            else:
                add_ppty = {}

            self.viewers[name] = Viewer(image_queue=image_queue,
                                        e_stop=e_stop,
                                        name=name,
                                        **add_ppty,
                                        **kwargs)


# =============================== MISC. Tools ================================


def get_last_from_queue(queue):
    """Function to empty queue to get last element from it.

    Return None if queue is initially empty, return last element otherwise.
    """
    element = None
    while True:
        try:
            element = queue.get(timeout=0)
        except Empty:
            break
    return element


def get_all_from_queue(queue):
    """Function to empty queue to get all elements from it as a list

    Return None if queue is initially empty, return last element otherwise.
    """
    elements = []
    while True:
        try:
            elements.append(queue.get(timeout=0))
        except Empty:
            break
    return elements


def max_possible_pixel_value(img):
    """Return max pixel value depending on image type, for use in plt.imshow.

    Input
    -----
    img: numpy array

    Output
    ------
    vmax: max pixel value (int or float or None)
    """
    if img.dtype == 'uint8':
        return 2**8 - 1
    elif img.dtype == 'uint16':
        return 2**16 - 1
    else:
        return None


class InfoSender(ABC):
    """Class to send information to display in Image Viewer.

    For example: fps, image number, etc.
    """
    def __init__(self, info_queue, e_stop, dt_check=1):
        """Parameters:

        - info_queue: queue into information is put
        - e_stop: stopping event (threading.Event or equivalent)
        - dt_check: how often (in seconds) information is sent
        """
        self.info_queue = info_queue
        self.e_stop = e_stop
        self.dt_check = dt_check

    @abstractmethod
    def _generate_info(self):
        """To be defined in subclass.

        Should return a str of info to print in the viewer.
        Should return a false-like value if no news info can be provided."""
        pass

    def run(self):
        """Send information periodically (blocking)."""

        while not self.e_stop.is_set():
            info = self._generate_info()
            if info:
                self.info_queue.put(info)
            else:
                self.info_queue.put('...')

            # dt_check should be long compared to the time required to
            # process and generate the info.
            self.e_stop.wait(self.dt_check)

    def start(self):
        """Same as run() but nonblocking."""
        Thread(target=self.run).start()


class LiveFpsCalculator(InfoSender):
    """"Calculate fps in real time from a queue supplying image times.

    fps values are sent back in another queue as str
    """

    def __init__(self, time_queue, info_queue, e_stop, dt_check=1):
        """Parameters:

        - time_queue: queue from which times arrive
        - info_queue: queue into which fps values are put
        - e_stop: stopping event (threading.Event or equivalent)
        - dt_check: how often (in seconds) times are checked to calculate fps
        """
        super().__init__(info_queue=info_queue, e_stop=e_stop, dt_check=dt_check)
        self.time_queue = time_queue

    @staticmethod
    def _calculate_fps(times):
        """Given an iterable of times, calculate average fps."""
        dt = np.diff(times).mean()
        return 1 / dt

    def _generate_info(self):
        """Calculate display fps from times put in time_queue."""
        times = get_all_from_queue(self.time_queue)
        if times:
            fps = self._calculate_fps(times)
            return f'{fps:.1f} fps'


# ----------------------------------------------------------------------------
# ============================== OpenCV viewers ==============================
# ----------------------------------------------------------------------------


class CvSingleViewer(SingleViewer):
    """Display camera images using OpenCV"""

    def _init_window(self):
        """Create window"""
        cv2.namedWindow(self.name, cv2.WINDOW_NORMAL)
        self.info = '...'

    def _manage_info(self, *args, **kwargs):
        """Redefined here because info needs to be printed on every image."""
        if self.info_queue:
            info = get_last_from_queue(self.info_queue)
            if info:
                self.info = info
            self._display_info(self.info, *args, **kwargs)

    def _update_window(self):
        """Indicate what happens at each step of the event loop."""
        data = get_last_from_queue(self.image_queue)
        if data is not None:

            image = self._measurement_to_image(data)
            self._manage_info(image=image)

            if image.ndim > 2:
                # openCV works with BGR data
                image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
            cv2.imshow(self.name, image)
            self._store_display_times()

    def _display_info(self, info, image=None):
        cv2.putText(image, info, (50, 50), cv2.FONT_HERSHEY_SIMPLEX,
                    1, (255, 255, 255), 2, cv2.LINE_AA)

    def run(self):
        """Loop to run live viewer"""
        while (cv2.getWindowProperty(self.name, cv2.WND_PROP_VISIBLE) > 0):
            self._update_window()
            if self.e_stop.is_set():
                cv2.destroyWindow(self.name)
                break
            cv2.waitKey(int(self.dt_graph * 1000))


class CvMultipleViewer(MultipleViewer):
    """Display several cameras at the same time using OpenCV"""

    def __init__(self,
                 image_queues,
                 e_stop=None,
                 dt_graph=0.01,
                 Viewer=CvSingleViewer,
                 **kwargs):
        """Parameters:

        - image_queues: dict {camera name: queue in which taken images are put.}
        - e_stop: stopping event (threading.Event or equivalent)
        - dt_graph: how often (in seconds) the GUI window is updated
        - Viewer: which Viewer to use (e.g. CvSingleViewer, CvStreamViewer, etc.)
        - **kwargs: any optional keyword arguments required by the Viewer.
        """
        super().__init__(image_queues=image_queues,
                         e_stop=e_stop,
                         dt_graph=dt_graph,
                         Viewer=Viewer,
                         **kwargs)

    def run(self):
        threads = []

        for viewer in self.viewers.values():
            threads.append(Thread(target=viewer.start))

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        cv2.destroyAllWindows()


# ----------------------------------------------------------------------------
# ============================ Matplotlib Viewers ============================
# ----------------------------------------------------------------------------


class MplAnimation:
    """Tools common to single-image and multiple-image animations"""

    def _update_figure(self, i):
        """Indicate what happens at each step of the matplotlib animation.

        Define in subclass.
        """
        pass

    def run(self):
        """Main function to run the animation"""
        ani = FuncAnimation(self.fig,
                            self._update_figure,
                            interval=int(self.dt_graph) * 1000,
                            blit=True,
                            cache_frame_data=False)
        plt.show(block=True)
        plt.close(self.fig)
        return ani

    def on_fig_close(self, event):
        """Anything to triggeer when figure is closed.

        A callback to on_fig_close must be declared in the subclass for this
        to be called.
        """

        # To be able to trigger on_stop() in a multiple image environment
        self.on_stop()


class MplSingleViewer(MplAnimation, SingleViewer):
    """Display camera images using Matplotlib"""

    def __init__(self, *args, dt_graph=0.04, ax=None, **kwargs):
        """Parameters:

        - dt_graph: interval (s) to update matplotlib animation
        - ax (optional): axes into which images are shown
        """
        super().__init__(*args, dt_graph=dt_graph, **kwargs)
        self.ax = ax

    def _init_window(self):

        if self.ax is None:
            self.fig, self.ax = plt.subplots()
        else:
            self.fig = self.ax.figure

            # To be able to trigger on_stop() in a multiple image environment
            self.fig.canvas.mpl_connect('close_event', self.on_fig_close)

        self._format_figure()

        self.init_done = False

    def _format_figure(self):
        """"Set colors, title etc."""
        self.fig.set_facecolor(bgcolor)
        self.ax.set_title(self.name, color=textcolor, fontfamily=fontfamily)

        for location in 'bottom', 'top', 'left', 'right':
            # self.ax.spines[location].set_color(textcolor)
            self.ax.spines[location].set_visible(False)

        self.ax.xaxis.label.set_color(textcolor)
        self.ax.tick_params(axis='both', colors=textcolor)

    def _init_image(self, image):
        self.im = self.ax.imshow(image,
                                 cmap='gray',
                                 animated=True,
                                 vmin=0,
                                 vmax=max_possible_pixel_value(image))
        self.init_done = True
        self.fig.tight_layout()
        self.xlabel = self.ax.set_xlabel('...', color=textcolor, fontfamily=fontfamily)

    def _update_figure(self, i):
        """Indicate what happens at each step of the matplotlib animation."""

        if self.e_stop.is_set():
            plt.close(self.fig)

        data = get_last_from_queue(self.image_queue)

        if data is not None:

            image = self._measurement_to_image(data)

            if not self.init_done:
                self._init_image(image)
            else:
                self.im.set_array(image)

            self._store_display_times()
            self._manage_info()

            return self.im,
        else:
            return ()

    def _display_info(self, info):
        self.xlabel.set_text(info)


class MplMultipleViewer(MplAnimation, MultipleViewer):
    """Display several cameras at the same time using Matplotlib"""

    def __init__(self,
                 image_queues,
                 e_stop=None,
                 dt_graph=0.04,
                 Viewer=MplSingleViewer,
                 **kwargs):
        """Parameters:

        - image_queues: dict {camera name: queue in which taken images are put.}
        - e_stop: stopping event (threading.Event or equivalent)
        - dt_graph: how often (in seconds) the GUI window is updated
        - Viewer: which Viewer to use (e.g. MplSingleViewer, MplStreamViewer, etc.)
        - **kwargs: any optional keyword arguments required by the Viewer.
        """
        self._create_axes(image_queues)

        super().__init__(image_queues=image_queues,
                         e_stop=e_stop,
                         dt_graph=dt_graph,
                         Viewer=Viewer,
                         add_ppties=self.axs,
                         add_ppty_name='ax',
                         **kwargs)

    def _create_axes(self, image_queues):
        """Generate figure/axes as a function of input names"""

        if len(image_queues) == 1:
            fig, ax = plt.subplots(1, 1, figsize=(10, 10))
            axes = ax,
        elif len(image_queues) == 2:
            fig, axes = plt.subplots(1, 2, figsize=(15, 8))
        else:
            raise Exception('Only 2 simultaneous cameras supported for now.')

        axs = {name: ax for name, ax in zip(image_queues, axes)}

        self.fig, self.axs = fig, axs

    def _init_window(self):
        for viewer in self.viewers.values():
            viewer._init_window()

    def _init_run(self):
        for viewer in self.viewers.values():
            viewer._init_run()

    def _update_figure(self, i):
        """Indicate what happens at each step of the matplotlib animation."""
        to_be_animated = ()
        for viewer in self.viewers.values():
            to_be_animated += viewer._update_figure(i)
        return to_be_animated


# ----------------------------------------------------------------------------
# ============================= Tkinter viewers ==============================
# ----------------------------------------------------------------------------


class TkSingleViewer(SingleViewer):
    """Live view of camera images using tkinter"""

    def __init__(self,
                 image_queue,
                 e_stop=None,
                 info_queue=None,
                 name="Camera",
                 dt_graph=0.01,
                 auto_size=True,
                 fit_to_screen=True,
                 root=None):
        """Parameters:

        - image_queue: queue in which taken images are put.
        - e_stop: stopping event (threading.Event or equivalent)
        - info_queue: if supplied, print info (received as str) below images
        - name: optional name for display purposes.
        - dt_graph: interval (s) to update Tkinter window
        - auto_size: autoscale image to window in real time
        - fit_to_screen: maximize window size when instantiated
        - root: Tkinter parent in which to display viewer (if not, tk.Tk())
        """
        super().__init__(image_queue=image_queue,
                         info_queue=info_queue,
                         e_stop=e_stop,
                         dt_graph=dt_graph,
                         name=name)

        self.auto_size = auto_size
        self.fit_to_screen = fit_to_screen
        self.root = tk.Tk() if root is None else root
        self.root.configure(bg=bgcolor)

    def _fit_to_screen(self):
        """Adapt window size to screen resolution/size"""
        w_screen = self.root.winfo_screenwidth()
        h_screen = self.root.winfo_screenheight()
        self.root.geometry(f"{0.9 * w_screen:.0f}x{0.9 * h_screen:.0f}")

    def _init_window(self):
        """Create tkinter window and elements."""

        if self.fit_to_screen:
            self._fit_to_screen()

        self.title_label = tk.Label(self.root, text=self.name,
                                    font=(fontfamily, 14),
                                    bg=bgcolor, fg=textcolor)

        self.title_label.pack(expand=True)

        self.image_label = tk.Label(self.root, highlightthickness=0)

        self.image_label.pack(expand=True)

        if self.info_queue:

            self.info_label = tk.Label(self.root,
                                       bg=bgcolor,
                                       fg=textcolor,
                                       font=(fontfamily, 12),
                                       text=str('...'))

            self.info_label.pack(expand=True)

    def _init_run(self):
        """If things need to be done before running (subclass if necessary)"""
        self.image_count = 0

    def run(self):
        """Main  loop for Tkinter GUI viewer"""
        self.update_window()
        self.root.mainloop()

    def _update_window(self):
        """"Update window, but without the after() method."""

        data = get_last_from_queue(self.image_queue)

        if data is not None:

            image = self._measurement_to_image(data)
            self.image_count += 1

            img = Image.fromarray(image)
            img_disp = self.prepare_displayed_image(img)

            self.img = ImageTk.PhotoImage(image=img_disp)
            self.image_label.configure(image=self.img)

            self._store_display_times()

            if self.info_queue:
                self.manage_info_queue()

    def update_window(self):
        """Update window, with the after() method."""

        self._update_window()

        if not self.e_stop.is_set():
            self.root.after(int(1000 * self.dt_graph), self.update_window)
        else:
            self.root.destroy()

    def prepare_displayed_image(self, img):
        """Resize image and/or calculate aspect ratio if necessary"""

        if self.image_count > 1:
            if self.auto_size:
                dimensions = self.adapt_image_to_window()
                try:
                    img_disp = img.resize(dimensions, Image.ANTIALIAS)
                except ValueError:  # somtimes dimensions are (0, 0) for some reason
                    img_disp = img
            else:
                img_disp = img

        else:  # Calculate aspect ratio on first image received
            self.aspect_ratio = img.height / img.width
            img_disp = img

        return img_disp

    def adapt_image_to_window(self):
        """Calculate new dimensions of image to accommodate window resizing."""

        window_width = self.root.winfo_width()
        window_height = self.root.winfo_height()

        target_width = 0.98 * window_width
        target_height = 0.85 * window_height

        target_ratio = target_height / target_width

        if target_ratio > self.aspect_ratio:
            width = int(target_width)
            height = int(target_width * self.aspect_ratio)
        else:
            height = int(target_height)
            width = int(target_height / self.aspect_ratio)

        return width, height

    def manage_info_queue(self):
        """Get info from queue and display it"""
        info = get_last_from_queue(self.info_queue)
        if info:
            self.info_label.config(text=info)


class TkMultipleViewer(MultipleViewer):
    """Live view of images from multiple cameras using tkinter"""

    def __init__(self,
                 image_queues,
                 e_stop=None,
                 dt_graph=0.01,
                 Viewer=TkSingleViewer,
                 **kwargs):
        """Parameters:

        - image_queues: dict {camera name: queue in which taken images are put.}
        - e_stop: stopping event (threading.Event or equivalent)
        - dt_graph: interval (s) to update Tkinter window
        - Viewer: which Viewer to use (e.g. TkSingleViewer, TkStreamViewer, etc.)
        - **kwargs: any optional keyword arguments required by the Viewer.
        """
        self.root = tk.Tk()
        self.root.configure(bg=bgcolor)

        frames = {name: tk.Frame(master=self.root) for name in image_queues}

        super().__init__(image_queues=image_queues,
                         e_stop=e_stop,
                         dt_graph=dt_graph,
                         Viewer=Viewer,
                         add_ppties=frames,
                         add_ppty_name='root',
                         fit_to_screen=False,  # Important to keep False
                         **kwargs)

        # How to place elements on window as a function of number of widgets
        dispositions = {1: (1, 1),
                        2: (1, 2),
                        3: (1, 3),
                        4: (2, 2)}

        n = len(self.image_queues)
        n1, n2 = dispositions[n]  # dimensions of grid to place elements
        positions = itertools.product(range(n1), range(n2))

        for viewer, position in zip(self.viewers.values(), positions):
            i, j = position
            viewer.root.grid(row=i, column=j, padx=5, pady=5, sticky='nsew')

        # Make columns and rows expand and be all the same size
        # Note: the str in uniform= is just an identifier
        # all columns / rows sharing the same string are kept of same size
        for i in range(n1):
            self.root.grid_rowconfigure(i, weight=1,
                                        uniform='same size rows')
        for j in range(n2):
            self.root.grid_columnconfigure(j, weight=1,
                                           uniform='same size columns')

    def _init_window(self):
        TkSingleViewer._fit_to_screen(self)
        for viewer in self.viewers.values():
            viewer._init_window()

    def _init_run(self):
        for viewer in self.viewers.values():
            viewer._init_run()

    def run(self):
        self.update_window()
        self.root.mainloop()

    def update_window(self):

        for viewer in self.viewers.values():
            viewer._update_window()

        if not self.e_stop.is_set():
            self.root.after(int(1000 * self.dt_graph), self.update_window)
        else:
            self.root.destroy()


# ============ For live streaming with real-time fps calculation =============


class CvStreamViewer(SingleStreamViewer, CvSingleViewer):
    pass


class MplStreamViewer(SingleStreamViewer, MplSingleViewer):
    pass


class TkStreamViewer(SingleStreamViewer, TkSingleViewer):
    pass


# ================ Integration into prevo.Record environments ================


class RecordInfoSender(InfoSender):
    """Class to send information to display in Image Viewer.

    For example: fps, image number, etc.
    """
    def __init__(self, recording, info_queue, e_stop, dt_check=0.2):
        """Parameters:

        - recording: object of subclass of prevo.RecordingBase
        - info_queue: queue into information is put
        - e_stop: stopping event (threading.Event or equivalent)
        - dt_check: how often (in seconds) information is sent
        """
        self.recording = recording
        super().__init__(info_queue=info_queue,
                         e_stop=e_stop,
                         dt_check=dt_check)

    def _generate_info(self):
        """To be defined in subclass.

        Should return a str of info to print in the viewer.
        Should return a false-like value if no news info can be provided."""
        return f'# {self.recording.num}'


class RecordSingleViewer:
    """Additional methods to SingleViewer and subclasses.

    For use in prevo.RecordBase environments."""

    def __init__(self, recordings, dt_check=0.2, **kwargs):

        super().__init__(info_queue=Queue(), **kwargs)

        self.recording = recordings[self.name]  # self.name defined in super()
        self.dt_check = dt_check

    def _init_run(self):
        super()._init_run()
        if self.info_queue:
            info_sender = RecordInfoSender(recording=self.recording,
                                           info_queue=self.info_queue,
                                           e_stop=self.e_stop,
                                           dt_check=self.dt_check)
            info_sender.start()

    def on_stop(self):
        """Here one does not want to stop recording when window is closed."""
        pass


class CvRecordSingleViewer(RecordSingleViewer, CvSingleViewer):
    pass


class MplRecordSingleViewer(RecordSingleViewer, MplSingleViewer):
    pass


class TkRecordSingleViewer(RecordSingleViewer, TkSingleViewer):
    pass


class RecordMultipleViewer:
    """Additional methods to MultipleViewers-type classes.

    For use in prevo.RecordBase environments."""

    def __init__(self, e_graph, **kwargs):
        super().__init__(Viewer=self.SingleViewer, **kwargs)
        self.e_graph = e_graph

    def on_stop(self):
        self.e_graph.clear()


class CvRecordMultipleViewer(RecordMultipleViewer, CvMultipleViewer):
    SingleViewer = CvRecordSingleViewer
    pass


class MplRecordMultipleViewer(RecordMultipleViewer, MplMultipleViewer):
    SingleViewer = MplRecordSingleViewer
    pass


class TkRecordMultipleViewer(RecordMultipleViewer, TkMultipleViewer):
    SingleViewer = TkRecordSingleViewer
    pass



"""Record several sensors as a function of time with interactive CLI and graph."""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>


# Standard library imports
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from threading import Event, Thread
from queue import Queue, Empty
from traceback import print_exc
import os

# Non-standard imports
from tqdm import tqdm
import oclock
from clivo import CommandLineInterface, ControlledProperty, ControlledEvent

# Local imports
from ..control import RecordingControl
from ..misc import mode_to_names

# ================================ MISC Tools ================================


def try_thread(function):
    """Decorator to make threaded functions return & print errors if errors occur"""
    def wrapper(*args, **kwargs):
        try:
            function(*args, **kwargs)
        except Exception:
            try:
                name = kwargs['name']
            except KeyError:
                sensor_info = ''
            else:
                sensor_info = f' for sensor "{name}"'
            try:
                nmax, _ = os.get_terminal_size()
            except OSError:
                nmax = 80
            print('\n')
            print('=' * nmax)
            print(f'ERROR in {function.__name__}() {sensor_info}')
            print('-' * nmax)
            print_exc()
            print('=' * nmax)
            print('\n')
            return
    return wrapper


# ========================== Sensor base classes =============================


class SensorError(Exception):
    pass


class SensorBase(ABC):
    """Abstract base class for sensor acquisition."""

    # If one wants to control specific properties of the sensor through
    # the CLI in Record or through pre-defined programs, it's possible
    # to specify them here, and they will be automatically added to
    # the controlled_properties attribute of the corresponding Recording
    # NOTE: this needs to be a class attribute and not an instance attribute,
    # because these properties are checked before the Sensor object is
    # instantiated.
    controlled_properties = ()

    def __init__(self):
        # (optional) : specific sensor errors to catch, can be an Exception
        # class or an iterable of exceptions; if not specified in subclass,
        # any exception is caught.
        self.exceptions = Exception

    def __enter__(self):
        """Context manager for sensor (enter). Optional."""
        return self

    def __exit__(self, exc_type, exc_val, traceback):
        """Context manager for sensor (exit). Optional."""
        pass

    @property
    @abstractmethod
    def name(self):
        """Name of sensor, Must be a class attribute."""
        pass

    def _read(self):
        """Read sensor.

        Either this method or _get_data() must be defined in subclasses.
        """
        with oclock.measure_time() as data:
            data['values'] = self._get_data()
        return data

    def _get_data(self):
        """Use this instead of subclassing _read() when you want the time
        of measurement and uncertainty on time to be added automatically.
        (useful when sensor does not return time information)"""
        return ()

    def read(self):
        """Read sensor and throw SensorError if measurement fails."""
        try:
            data = self._read()
        except self.exceptions:
            raise SensorError(f'Impossible to read {self.name} sensor')
        else:
            return data


# =============== Default controlled properties for recordings ===============

timer_ppty = ControlledProperty(attribute='timer.interval',
                                readable='Δt (s)',
                                commands=('dt',))

active_ppty = ControlledProperty(attribute='active',
                                 readable='Rec. ON',
                                 commands=('on',))

# ========================== Recording base class ============================


class RecordingBase(ABC):
    """Base class for recording object used by RecordBase. To subclass"""

    def __init__(self,
                 Sensor,
                 path='.',
                 dt=1,
                 ctrl_ppties=(),
                 active=True,
                 continuous=False,
                 warnings=False,
                 precise=False,
                 programs=(),
                 control_params=None):
        """Init Recording object.

        Parameters
        ----------

        - Sensor: subclass of SensorBase.
        - path: directory in which data is recorded.
        - dt: time interval between readings (default 1s).
        - ctrl_ppties: optional iterable of properties (ControlledProperty
                       objects) to control on the recording in addition to
                       default ones (time interval and active on/off)
        - active: if False, do not record data until self.active set to True.
        - continuous: if True, take data as fast as possible from sensor.
        - warnings: if True, print warnings of Timer (e.g. loop too short).
        - precise: if True, use precise timer in oclock (see oclock.Timer).
        - programs: iterable of programs, which are object of the
                    prevo.control.Program class or subclasses.
                    --> optional pre-defined temporal pattern of change of
                    properties of recording (e.g. define some times during
                    which sensor is active or not, or change time interval
                    between data points after some time, etc.)
        - control_params: dict {command: kwargs} containing any kwargs to pass
                          to the program controls (e.g. dt, range_limits, etc.)
                          Note: if None, use default params
                          (see RecordingControl class)
        """
        self.Sensor = Sensor
        self.name = Sensor.name
        self.timer = oclock.Timer(interval=dt,
                                  name=self.name,
                                  warnings=warnings,
                                  precise=precise)
        self.path = Path(path)

        self._active = active  # can be set to False to temporarily stop recording from sensor
        self.continuous = continuous

        # To be defined in subclass.
        # Object that manages how data is written to files.
        # Must have a .file attribute whith is a pathlib.Path object, and
        # .init_file() method to create / init the file when recording starts
        self.file_manager = None

        # Iterable of the recording properties that the program / CLI control.
        # Possibility to add other properties in subclasses
        # (need to be of type ControlledProperty or subclass)
        if self.continuous:
            # If sensor instantiated as continuous (e.g. streaming cameras)
            # we assume that the timer is internally controlled in the sensor
            # so that it does not make sense to control the time interval
            # from the CLI.
            # If one has a sensor when one wants to switch from continuous
            # to not continuous, it's better to instantiate it as not
            # continuous (possibly with on=False) and then switch to
            # continuous and on=True.
            default_ppties = (active_ppty,)
        else:
            default_ppties = (timer_ppty, active_ppty)
        self.controlled_properties = default_ppties + ctrl_ppties
        self._add_sensor_controlled_properties()
        self._generate_ppty_dict()

        # Optional temporal programs to make controlled properties evolve.
        self._init_programs(programs=programs,
                            control_params=control_params)

    def __repr__(self):
        return f'{self.__class__.__name__} ({self.name})'

    @property
    def active(self):
        """In case a bunch of things (e.g. sensor start/stop) need to be done
        when changing the active property --> subclass this"""
        return self._active

    @active.setter
    def active(self, value):
        self._active = value

    # Private methods --------------------------------------------------------

    def _add_sensor_controlled_properties(self):
        for ppty in self.Sensor.controlled_properties:
            new_attribute = 'sensor.' + ppty.attribute
            new_ppty = ControlledProperty(attribute=new_attribute,
                                          readable=ppty.readable,
                                          commands=ppty.commands)
            self.controlled_properties += (new_ppty,)

    def _generate_ppty_dict(self):
        """Associate property commands to property objects"""
        self.ppty_commands = {}
        for ppty in self.controlled_properties:
            for ppty_cmd in ppty.commands:
                self.ppty_commands[ppty_cmd] = ppty

    def _init_programs(self, programs, control_params):
        self.programs = ()
        for supplied_program in programs:

            # In case same program is supplied to different recordings,
            # because a program object cannot be run several times in parallel.
            program = supplied_program.copy()

            ppty_cmd = program.quantity
            ppty = self.ppty_commands[ppty_cmd]
            if control_params is not None:
                control_kwargs = control_params.get(ppty_cmd, {})
            else:
                control_kwargs = {}
            log_filename = f'Control_Log_{self.name}_[{ppty.readable}].txt'
            program.control = RecordingControl(recording=self,
                                               ppty=ppty,
                                               log_file=log_filename,
                                               savepath=self.path,
                                               **control_kwargs)
            self.programs += (program,)

    def _stop_programs(self):
        for program in self.programs:
            program.stop()

    def _set_property(self, ppty, value):
        """Set property of recording.

        recording: RecordingBase object or subclass on which property is applied
        ppty: ControlledProperty object or subclass.
        value: value of property to apply.
        """
        if ppty not in self.controlled_properties:
            return
        try:
            # avoids having to pass a convert function
            exec(f'self.{ppty.attribute} = {value}')
        except Exception as e:
            print(f"WARNING: Could not set {ppty.readable} for {self.name} to {value}.\n Exception: {e}")

    # Compulsory methods / properties to subclass ----------------------------

    @abstractmethod
    def save(self, measurement, file):
        """Write data of measurement to (already open) file.

        file is the file object yielded by the open() context manager.
        """
        pass

    # Optional methods to subclass -------------------------------------------

    def format_measurement(self, data):
        """How to format the data given by self.Sensor.read().

        Returns a measurement object (e.g. dict, value, custom class etc.)."""
        return data

    def after_measurement(self):
        """Define what to do after measurement has been done and formatted.

        Acts on the recording object but does not return anything.
        (Optional)
        """
        pass

    # General methods and attributes (can be subclassed if necessary) --------

    def print_info_on_failed_reading(self, status):
        """Displays relevant info when reading fails."""
        t_str = datetime.now().isoformat(sep=' ', timespec='seconds')
        if status == 'failed':
            print(f'{self.name} reading failed ({t_str}). Retrying ...')
        elif status == 'resumed':
            print(f'{self.name} reading resumed ({t_str}).')

    def on_stop(self):
        """What happens when a stop event is requested in the CLI"""
        self.timer.stop()
        self._stop_programs()

class RecordBase:
    """Asynchronous recording of several Recordings object.

    Recordings objects are of type RecordingBase."""

    # Warnings when queue size goes over some limits
    queue_warning_limits = 1e1, 1e2, 1e3, 1e4, 1e5, 1e6, 1e7, 1e8, 1e9

    def __init__(self,
                 recordings,
                 path='.',
                 on_start=(),
                 dt_save=1.9,
                 dt_request=0.7,
                 dt_check=1.3):
        """Init base class for recording data

        Parameters
        ----------
        - recordings: iterable of recording objects (RecordingBase or subclass)
        - path: directory in which data is recorded.
        - on_start: optional iterable of objects with a .start() or .run()
                    method, that need to be started at the same time as
                    Record.start().
                    Note: for now, start() and run() need to be non-blocking.
        - dt_save: how often (in seconds) queues are checked and written to files
                   (it is also how often files are open/closed)
        - dt_request: time interval (in seconds) for checking user requests
                      (e.g. graph pop-up)
        - dt_check: time interval (in seconds) for checking queue sizes.
        """
        self.recordings = {rec.name: rec for rec in recordings}
        self.create_events()

        self.path = Path(path)
        self.path.mkdir(exist_ok=True)
        self.on_start = on_start

        self.dt_save = dt_save
        self.dt_request = dt_request
        self.dt_check = dt_check

        # Data is stored in a queue before being saved
        self.q_save = {name: Queue() for name in self.recordings}

        # Data queue for plotting when required
        self.q_plot = {name: Queue() for name in self.recordings}

        # Any additional functions that need to be run along the other threads
        # (to be defined in subclasses)
        self.additional_threads = []

    # =========== Optional methods and attributes for subclassing ============

    def save_metadata(self):
        """Save experimental metadata. To be defined in subclass"""
        pass

    def data_plot(self):
        """What to do with data when graph event is triggered"""
        pass

    # ============================= Misc methods =============================

    @staticmethod
    def increment_filename(file):
        """Find an increment on file name, e.g. -1, -2 etc. to create file
        that does not exist.

        Convenient for some uses, e.g. not overwrite metadata file, etc.
        """
        full_name_str = str(file.absolute())
        success = False
        n = 0
        while not success:
            n += 1
            new_stem = f'{file.stem}-{n}'
            new_name = full_name_str.replace(file.stem, new_stem)
            new_file = Path(new_name)
            if not new_file.exists():
                success = True
        return new_file

    # ------------------------------------------------------------------------
    # ============================= INIT METHODS =============================
    # ------------------------------------------------------------------------

    def create_events(self):
        """Create event objects managed by the CLI"""
        self.e_stop = Event()  # event set to stop recording when needed.
        self.e_graph = Event()  # event set to start plotting the data in real time

        graph_event = ControlledEvent(event=self.e_graph,
                                      readable='graph',
                                      commands=('g', 'graph'))

        stop_event = ControlledEvent(event=self.e_stop,
                                     readable='stop',
                                     commands=('q', 'Q', 'quit'))

        self.events = graph_event, stop_event

    def parse_initial_user_commands(self, ppty_kwargs):
        """Check if user input contains specific properties.

        The values to apply for these properties are stored in a dict and
        will be applied to each recording when launched.

        If generic input (e.g. 'dt=10'), set all recordings to that value
        If specific input (e.g. dt_P=10), update recording to that value
        """
        self.initial_ppty_settings = {name: {} for name in self.recordings}

        global_commands = {}
        specific_commands = {}

        for cmd, value in ppty_kwargs.items():
            try:
                ppty_cmd, name = cmd.split('_', maxsplit=1)  # e.g. dt_P = 10
            except ValueError:                               # e.g. dt=10
                global_commands[cmd] = value
            else:
                specific_commands[name] = ppty_cmd, value

        # Apply first global values to all recordings ------------------------
        for ppty_cmd, value in global_commands.items():
            for name, recording in self.recordings.items():
                ppty = recording.ppty_commands[ppty_cmd]
                self.initial_ppty_settings[name][ppty] = value

        # Then apply commands to specific recordings if specified ------------
        for name, (ppty_cmd, value) in specific_commands.items():
            ppty = self.recordings[name].ppty_commands[ppty_cmd]
            self.initial_ppty_settings[name][ppty] = value

    # ------------------------------------------------------------------------
    # =================== START RECORDING (MULTITHREAD) ======================
    # ------------------------------------------------------------------------

    def add_named_threads(self, function):
        for name in self.recordings:
            kwargs = {'name': name}
            self.threads.append(Thread(target=function, kwargs=kwargs))

    def add_other_threads(self):
        """Add other threads for additional functions defined by user."""
        for func in self.additional_threads:
            self.threads.append(Thread(target=func))

    def start_supplied_objects(self):
        """Start user-supplied non-blocking objects with a start() or run() method."""
        for obj in self.on_start:
            for method in 'run', 'start':
                try:
                    getattr(obj, method)()
                except AttributeError:
                    pass
                else:
                    break
            else:
                print(f'WARNING - {obj.__class__.__name} does not have a'
                      f'run() or start() method --> not started. ')

    def start(self, **ppty_kwargs):
        """Start recording.

        Parameters
        ----------
        - ppty_kwargs: optional initial setting of properties.
                     (example dt=10 for changing all time intervals to 10
                      or dt_P=60 to change only time interval of recording 'P')
        """
        # Check if user inputs particular initial settings for recordings
        self.parse_initial_user_commands(ppty_kwargs)

        self.save_metadata()

        print(f'Recording started in folder {self.path.resolve()}')

        error_occurred = False

        try:

            self.threads = []

            self.add_named_threads(self.data_read)
            self.add_named_threads(self.data_save)
            self.threads.append(Thread(target=self.check_queue_sizes))
            self.add_other_threads()

            for thread in self.threads:
                thread.start()

            # Add CLI. This one is a bit particular because it is blocking
            # with input() and has to be manually stopped. ----------------
            self.cli_thread = Thread(target=self.cli)
            self.cli_thread.start()

            # Start any other user-supplied objects (need to be nonblocking)
            self.start_supplied_objects()

            # real time graph (triggered by CLI, runs in main thread due to
            # GUI backend problems if not) --------------------------------
            self.data_graph()

        except Exception as e:
            error_occurred = True
            print(f'\nERROR during recording: {e}. \n Stopping ... \n')
            print_exc()

        except KeyboardInterrupt:
            error_occurred = True
            print('\nManual interrupt by Ctrl-C event.\n Stopping ...\n')

        finally:

            self.e_stop.set()

            for thread in self.threads:
                thread.join()

            for obj in self.on_start:
                obj.stop()

            if error_occurred:
                print('\nIMPORTANT: CLI still running. Input "q" to stop.\n')

            self.cli_thread.join()

            print('Recording Stopped')

    # ------------------------------------------------------------------------
    # =============================== Threads ================================
    # -------------------- (CLI is defined elsewhere) ------------------------

    # =========================== Data acquisition ===========================

    @try_thread
    def cli(self):
        if self.recordings:  # if no recordings provided, no need to record.
            command_input = CommandLineInterface(self.recordings, self.events)
            command_input.run()
        else:
            self.e_stop.set()
            raise ValueError('No recordings provided. Stopping ...')

    @try_thread
    def data_read(self, name):
        """Read data from sensor and store it in data queues."""

        # Init ---------------------------------------------------------------

        recording = self.recordings[name]
        saving_queue = self.q_save[name]
        plotting_queue = self.q_plot[name]
        failed_reading = False  # True temporarily if P or T reading fails

        # Recording loop -----------------------------------------------------

        with recording.Sensor() as sensor:

            recording.sensor = sensor

            # Initial setting of properties is done here in case one of the
            # properties acts on the sensor object, which is not defined
            # before this point.
            # Note: (_set_property() does not do anything if the property
            # does not exist for the recording of interest)
            for ppty, value in self.initial_ppty_settings[name].items():
                recording._set_property(ppty=ppty, value=value)

            # Optional pre-defined program for change of recording properties
            for program in recording.programs:
                program.run()

            # Without this here, the first data points are irregularly spaced.
            recording.timer.reset()

            while not self.e_stop.is_set():

                if not recording.active:
                    if not recording.continuous:
                        # to avoid checking too frequently if active or not.
                        recording.timer.checkpt()
                    continue

                try:
                    data = sensor.read()

                # Measurement has failed .........................................
                except SensorError:
                    if not failed_reading:  # means it has not failed just before
                        recording.print_info_on_failed_reading('failed')
                    failed_reading = True

                # Measurement is OK ..............................................
                else:
                    if failed_reading:      # means it has failed just before
                        recording.print_info_on_failed_reading('resumed')
                        failed_reading = False

                    measurement = recording.format_measurement(data)
                    recording.after_measurement()

                    # Store recorded data in a first queue for saving to file
                    saving_queue.put(measurement)

                    # Store recorded data in another queue for plotting
                    if self.e_graph.is_set():
                        plotting_queue.put(measurement)

                # Below, this means that one does not try to acquire data right
                # away after a fail, but one waits for the usual time interval
                finally:
                    if not recording.continuous:
                        recording.timer.checkpt()

            else:
                recording.on_stop()

    # ========================== Write data to file ==========================

    @try_thread
    def data_save(self, name):
        """Save data that is stored in a queue by data_read."""

        recording = self.recordings[name]
        saving_queue = self.q_save[name]
        saving_timer = oclock.Timer(interval=self.dt_save)

        recording.file_manager.init_file()

        while not self.e_stop.is_set():

            # Open and close file at each cycle to be able to save periodically
            # and for other users/programs to access the data simultaneously
            with open(recording.file_manager.file, 'a', encoding='utf8') as file:

                while not saving_timer.interval_exceeded:

                    try:
                        measurement = saving_queue.get(timeout=self.dt_save)
                    except Empty:
                        pass
                    else:
                        if measurement is not None:
                            recording.save(measurement, file)

                    if self.e_stop.is_set():  # Move to buffering waitbar
                        break

                # periodic check whether there is data to save
                saving_timer.checkpt()

        # Buffering waitbar --------------------------------------------------

        if not saving_queue.qsize():
            return

        print(f'Data buffer saving started for {name}')

        # The nested statements below, similarly to above, ensure that
        # recording.file_manager.file is opened and closed regularly to avoid
        # loosing too much data if there is an error.

        with tqdm(total=saving_queue.qsize()) as pbar:
            while True:
                try:
                    with open(recording.file_manager.file, 'a', encoding='utf8') as file:
                        saving_timer.reset()
                        while not saving_timer.interval_exceeded:
                            measurement = saving_queue.get(timeout=self.dt_save)
                            self._try_save(measurement, recording, file)
                            pbar.update()
                except Empty:
                    break

        print(f'Data buffer saving finished for {name}')

    # =========================== Real-time graph ============================

    def data_graph(self):
        """Manage requests of real-time plotting of data during recording."""

        while not self.e_stop.is_set():

            if self.e_graph.is_set():
                # no need to reset e_graph here, because data_plot is blocking in
                # this version of the code (because of the plt.show() and
                # FuncAnimation). If data_plot is changed, it might be useful to
                # put back a e_graph.clear() here.

                self.data_plot()

            self.e_stop.wait(self.dt_request)  # check whether there is a graph request

    # ========================== Check queue sizes ===========================

    def _init_queue_size_info(self):
        """Create dict indicating that no queue limit is over among limits"""
        queue_size_over = {}
        for name in self.recordings:
            queue_size_over[name] = {limit: False
                                     for limit in self.queue_warning_limits}
        return queue_size_over

    def _check_queue_size(self, name, q, q_size_over, q_type):
        """Check that queue does not go beyond specified limits"""
        for qmax in self.queue_warning_limits:

            if q.qsize() > qmax:
                if not q_size_over[qmax]:
                    print(f'\nWARNING: {q_type} buffer size for {name} over {qmax} elements')
                    q_size_over[qmax] = True

            if q.qsize() <= qmax:
                if q_size_over[qmax]:
                    print(f'\n{q_type} buffer size now below {qmax} for {name}')
                    q_size_over[qmax] = False

    @try_thread
    def check_queue_sizes(self):
        """Periodically verify that queue sizes are not over limits"""

        self.q_save_size_over = self._init_queue_size_info()
        self.q_plot_size_over = self._init_queue_size_info()

        while not self.e_stop.is_set():

            for name in self.recordings:

                self._check_queue_size(name=name,
                                       q=self.q_save[name],
                                       q_size_over=self.q_save_size_over[name],
                                       q_type='Saving')

                if self.e_graph.is_set():

                    self._check_queue_size(name=name,
                                           q=self.q_save[name],
                                           q_size_over=self.q_save_size_over[name],
                                           q_type='Plotting')

            self.e_stop.wait(self.dt_check)

    @classmethod
    def create(cls,
               mode=None,
               sensors=(),
               default_names=None,
               recording_types=None,
               recording_kwargs=None,
               programs=None,
               control_params=None,
               path='.',
               **kwargs):
        """Factory method to generate a Record object from sensor names etc.

        Parameters
        ----------
        - mode (default: all sensors): sensor names in any order (e.g. 'PTB1')
               and potentially with separators (e.g. 'P-T-B1')
        - sensors: iterable of all sensor classes [Sensor1, Sensor2, etc.]
        - default_names (optional): default sensor names to record if mode
                                    is not supplied (left to None)
        - recording_types: dict {sensor name: Recording} indicating what class
                           of recording (RecordingBase class or subclass)
                           needs to be instantiated for each sensor.
        - recording_kwargs: dict {sensor name: kwargs}
        - programs: dict {mode: programs} with mode same as the mode argument above
                    (describing all sensor recordings that are concerned by the
                    supplied programs), and programs an iterable of objects from
                    the prevo.control.Program class or subclass.
        - control_params: dict {command: kwargs} containing any kwargs to pass
                          to the program controls (e.g. dt, range_limits, etc.)
                          Note: if None, use default params
                          (see RecordingControl class)
        - **kwargs is any keyword arguments to pass to Record __init__
          (including dt_save, ppty_kwargs etc.)
        """
        all_sensors = {Sensor.name: Sensor for Sensor in sensors}
        name_info = {'possible_names': all_sensors,
                     'default_names': default_names}
        names = mode_to_names(mode=mode, **name_info)

        programs = {} if programs is None else programs

        all_programs = {name: () for name in names}
        for pgm_mode, programs in programs.items():
            pgm_names = mode_to_names(mode=pgm_mode, **name_info)
            for name in pgm_names:
                all_programs[name] += programs

        recordings = []
        for name in names:
            Recording = recording_types[name]
            recording = Recording(Sensor=all_sensors[name],
                                  path=path,
                                  programs=all_programs[name],
                                  control_params=control_params,
                                  **recording_kwargs[name])
            recordings.append(recording)

        return cls(recordings=recordings, path=path, **kwargs)



"""General file input/output with csv files"""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>


# Standard library
from pathlib import Path

# Nonstandard
try:
    import pandas as pd
except ModuleNotFoundError:
    pass


class CsvFile:

    def __init__(self,
                 filename,
                 column_names=None,
                 column_formats=None,
                 path='.',
                 csv_separator='\t'):
        """Parameters:

        - path: str or pathlib.Path object: folder in which file is located
        - filename: name of file within path folder.
        - csv_separator: separator (str) used to separate data in file
        - column_names (optional, for saving data): iterable of column names
        - column_formats (optional): iterable of str formatings of data in columns
        """
        self.path = Path(path)
        self.file = self.path / filename
        self.csv_separator = csv_separator
        self.column_names = column_names
        self.column_formats = column_formats

        if column_formats is None and self.column_names is not None:
            self.column_formats = ('',) * len(column_names)

    def load(self, nrange=None):
        """Load data recorded in path, possibly with a range of indices (n1, n2).

        Input
        -----
        - nrange: select part of the data:
            - if nrange is None (default), load the whole file.
            - if nrange = (n1, n2), loads the file from line n1 to line n2,
              both n1 and n2 being included (first line of data is n=1).

        Output
        ------
        Pandas DataFrame of the requested size.
        """
        if nrange is None:
            kwargs = {}
        else:
            n1, n2 = nrange
            kwargs = {'skiprows': range(1, n1),
                      'nrows': n2 - n1 + 1}

        return pd.read_csv(self.file, delimiter=self.csv_separator, **kwargs)

    def number_of_lines(self):
        """Return number of lines of a file"""
        with open(self.file, 'r') as f:
            for i, line in enumerate(f):
                pass
            try:
                return i + 1
            except UnboundLocalError:  # handles the case of an empty file
                return 0

    def number_of_measurements(self):
        """Can be subclassed (here, assumes column titles)"""
        return self.number_of_lines() - 1

    # ---------- Methods that work on already opened file managers -----------

    def _init_file(self, file):
        """What to do with file when recording is started."""
        # Line below allows the user to re-start the recording and append data
        if self.number_of_lines() == 0:
            self._write_columns(file)

    def _write_columns(self, file):
        """How to init the file containing the data (when file already open)"""
        if self.column_names is None:
            return
        columns_str = f'{self.csv_separator.join(self.column_names)}\n'
        file.write(columns_str)

    def _write_line(self, data, file):
        """Save data to file when file is already open."""
        data_str = [f'{x:{fmt}}' for x, fmt in zip(data, self.column_formats)]
        line_for_saving = self.csv_separator.join(data_str) + '\n'
        file.write(line_for_saving)

    # ----------- Corresponding methods that open the file manager -----------

    def init_file(self):
        """What to do with file when recording is started."""
        with open(self.file, 'a', encoding='utf8') as file:
            self._init_file(file)



"""Record several sensors as a function of time with interactive CLI and graph."""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>


import gittools
import prevo

# Local imports
from .general import RecordingBase, RecordBase
from ..csv import CsvFile
from ..viewers import CvRecordMultipleViewer
from ..viewers import MplRecordMultipleViewer
from ..viewers import TkRecordMultipleViewer

# Optional, nonstandard
try:
    from PIL import Image
except ModuleNotFoundError:
    pass


class ImageRecording(RecordingBase):
    """Recording class to record images and associated timestamps."""

    def __init__(self,
                 Sensor,
                 timestamp_filename,
                 path='.',
                 image_path=None,
                 extension=None,
                 ndigits=5,
                 quality=None,
                 column_names=None,
                 column_formats=None,
                 csv_separator='\t',
                 **kwargs):
        """Init ImageRecording object.

        Parameters
        ----------
        - Sensor: subclass of SensorBase.
        - timestamp_filename: name of csv/tsv file in which timestamp data
                              will be saved.
        - path: directory in which csv/tsv file will be created.
        - image_path: path in which individual images will be recorded. By
                      default, creates a folder with the name of the sensor
                      in the 'path' directory defined above.
        - extension: e.g. '.tif', '.jpg', etc. of the saved images (None: default)
        - ndigits: number of digits for the image counter in the filename
        - quality: for compressed image formats (e.g. jpg, tif), see
        https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html
        - column_names: iterable of str (name of columns in csv file)
        - column_formats: iterable of str (optional, str formatting)
        - csv_separator: character to separate columns in CSV file.

        Additional kwargs from RecordingBase:
        - dt: time interval between readings (default 1s).
        - ctrl_ppties: optional iterable of properties (ControlledProperty
                       objects) to control on the recording in addition to
                       default ones (time interval and active on/off)
        - active: if False, do not record data until self.active set to True.
        - continuous: if True, take data as fast as possible from sensor.
        - warnings: if True, print warnings of Timer (e.g. loop too short).
        - precise: if True, use precise timer in oclock (see oclock.Timer).
        - programs: iterable of programs, which are object of the
                    prevo.control.Program class or subclasses.
                    --> optional pre-defined temporal pattern of change of
                    properties of recording (e.g. define some times during
                    which sensor is active or not, or change time interval
                    between data points after some time, etc.)
        - control_params: dict {command: kwargs} containing any kwargs to pass
                          to the program controls (e.g. dt, range_limits, etc.)
                          Note: if None, use default params
                          (see RecordingControl class)
        """
        super().__init__(Sensor=Sensor, path=path, **kwargs)

        # Here, file manager manages only the timestamp file, not the images
        self.file_manager = CsvFile(filename=timestamp_filename,
                                    column_names=column_names,
                                    column_formats=column_formats,
                                    path=path,
                                    csv_separator=csv_separator)

        self.image_path = self.path / Sensor.name if image_path is None else image_path
        self.image_path.mkdir(exist_ok=True)

        self.column_names = column_names

        if extension is None:
            # because for fast recording, .tif saving is much faster than png
            self.extension = '.tif' if self.continuous else '.png'
        else:
            self.extension = extension

        self.quality = quality
        self.ndigits = ndigits
        self.fmt = f'0{self.ndigits}'

        # number of images already recorded when record is called
        # (e.g. due to previous recording interrupted and restared)
        # The with open creates the file if not exists yet.
        with open(self.file_manager.file, 'a', encoding='utf8'):
            n_lines = self.file_manager.number_of_lines()
        self.num = n_lines - 1 if n_lines > 1 else 0

    def format_measurement(self, data):
        """How to format the data"""
        return {'name': self.name, 'num': self.num, **data}

    def after_measurement(self):
        """What to do after formatting data."""
        self.num += 1  # update number of recorded images for specific sensor

    def _generate_image_filename(self, measurement):
        """How to name images. Can be subclassed."""
        basename = f"{self.name}-{measurement['num']:{self.fmt}}"
        return basename + self.extension

    def _save_image(self, measurement, file):
        """How to save images to individual files. Can be subclassed."""
        img = measurement['image']
        if self.quality is None:
            Image.fromarray(img).save(file)
        else:
            Image.fromarray(img).save(file, quality=self.quality)

    def _save_timestamp(self, measurement, file):
        """How to save timestamps and other info to (opened) timestamp file"""
        filename = self._generate_image_filename(measurement)
        info = {'filename': filename, **measurement}
        data = [info[x] for x in self.column_names]
        self.file_manager._write_line(data, file)

    def save(self, measurement, file):
        """Write data to .tsv file with format: datetime / delta_t / value(s).

        Input
        -----
        - data: dict of data from the read() function
        - file: file in which to save the data
        """
        img_filename = self._generate_image_filename(measurement)
        self._save_image(measurement, file=self.image_path / img_filename)
        self._save_timestamp(measurement, file=file)

    @property
    def info(self):
        """Additional metadata info to save in metadata file. Subclass."""
        return {}


class ImageRecord(RecordBase):
    """Main class managing simultaneous temporal recordings of images."""

    def __init__(self,
                 recordings,
                 metadata_filename='Images_Metadata.json',
                 checked_modules=(),
                 dt_graph=0.1,
                 viewer='tk',
                 dirty_ok=True,
                 **kwargs):
        """Init ImageRecord object.

        Parameters
        ----------
        - recordings: iterable of recording objects (RecordingBase or subclass)
        - metadata_filename: name of .json file in which metadata is saved
        - checked_modules: iterable of python modules, the version of which
                           will be saved in metadata, in addition to prevo.
        - dt_graph: time interval to refresh numerical graph.
        - viewer: type of viewer to display images ('mpl', 'tk' or 'cv)
        - dirty_ok: if False, record cannot be started if git repositories are
                    not clean (commited).

        Additional kwargs from RecordingBase:
        - path: directory in which data is recorded.
        - on_start: optional iterable of objects with a .start() or .run()
                    method, that need to be started at the same time as
                    Record.start().
                    Note: for now, start() and run() need to be non-blocking.
        - dt_save: how often (in seconds) queues are checked and written to files
                   (it is also how often files are open/closed)
        - dt_request: time interval (in seconds) for checking user requests
                      (e.g. graph pop-up)
        - dt_check: time interval (in seconds) for checking queue sizes.
        """
        super().__init__(recordings=recordings, **kwargs)
        self.metadata_filename = metadata_filename
        self.checked_modules = set((prevo,) + tuple(checked_modules))

        # Viewing options -------------
        self.dt_graph = dt_graph
        self.viewer = viewer

        self.dirty_ok = dirty_ok
        self.get_image_recordings()

    def get_image_recordings(self):
        """Useful when combined with other recording types (e.g. vacuum)"""
        self.image_recordings = {}
        for name, recording in self.recordings.items():
            if issubclass(recording.__class__, ImageRecording):
                self.image_recordings[name] = recording

    def _save_metadata(self, filename):
        """To be able to call save_metadata() with arbitrary filenames"""
        metadata_file = self.path / filename

        if metadata_file.exists():
            metadata_file = self.increment_filename(metadata_file)

        info = {}

        for name, recording in self.image_recordings.items():

            info[name] = {
                'sensor': repr(recording.Sensor),
                'initial image number': recording.num,
                'extension': recording.extension,
                'digit number': recording.ndigits,
                'quality': str(recording.quality),
                **recording.info,
            }

        gittools.save_metadata(metadata_file,
                               info=info,
                               module=self.checked_modules,
                               dirty_warning=True,
                               dirty_ok=self.dirty_ok,
                               notag_warning=True,
                               nogit_ok=True,
                               nogit_warning=True)

    def save_metadata(self):
        """Save sensor info and version of modules used"""
        self._save_metadata(filename=self.metadata_filename)

    def data_plot(self):
        """What to do when graph event is triggered"""

        Viewers = {'cv': CvRecordMultipleViewer,
                   'mpl': MplRecordMultipleViewer,
                   'tk': TkRecordMultipleViewer}

        # In case the queue contains other measurements than images
        # (e.g. numerical data from temperature sensors, etc.)
        image_queues = {name: self.q_plot[name]
                        for name in self.image_recordings}

        kwargs = {'recordings': self.image_recordings,
                  'image_queues': image_queues,
                  'e_stop': self.e_stop,
                  'e_graph': self.e_graph,
                  'dt_graph': self.dt_graph}

        Viewers[self.viewer](**kwargs).start()



"""General tools and base classes for the prevo.plot module."""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>


# Standard library imports
from abc import ABC, abstractmethod
from queue import Empty

# Non standard imports
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from matplotlib import cm


# ----------------------------------------------------------------------------


class GraphBase(ABC):
    """Base class for managing plotting of arbitrary measurement data"""

    @abstractmethod
    def format_measurement(self, measurement):
        """Transform measurement from the queue into something usable by manage_data()

        Can be subclassed to adapt to various applications.
        Here, assumes data is incoming in the form of a dictionary with at
        least keys:
        - 'name' (str, identifier of sensor)
        - 'time (unix)' (floar or array of floats)
        - 'values' (iterable of values, or iterable of arrays of values)

        Subclass to adapt to applications.
        """
        pass

    def _plot(self, data):
        """Plot individual measurement on existing graph.

        Uses output of self.format_measurement() (data)
        To subclass.
        """
        pass

    def plot(self, measurement):
        """Wrapper around _plot()."""
        # The line below allows some sensors to avoid being plotted by reading
        # None when called.
        if measurement is None:
            return

        data = self.format_measurement(measurement)
        self._plot(data)

    @property
    def animated_artists(self):
        """Optional property to define for graphs updated with blitting."""
        return ()


class NumericalGraphBase(GraphBase):

    def __init__(self,
                 names,
                 data_types,
                 colors=None,
                 legends=None,
                 linestyles=None,
                 linestyle='.',
                 data_as_array=False):
        """Initiate figures and axes for data plot as a function of asked types.

        Input
        -----
        - names: iterable of names of recordings/sensors that will be plotted.
        - data types: dict with the recording names as keys, and the
                      corresponding data types as values.
                      (dict can have more keys than those in 'names')
        - colors: optional dict of colors with keys 'fig', 'ax', and the
                    names of the recordings.
        - legends: optional dict of legend names (iterable) corresponding to
                   all channels of each sensor, with the names of the
                   recordings as keys.
        - linestyles: optional dict of linestyles (iterable) to distinguish
                      channels and sensors, with the names of the recordings
                      as keys. If not specified (None), all lines have the
                      linestyle defined by the `linestyle=` parameter (see
                      below). If only some recordings are specified, the other
                      recordings have the default linestyle or the linestyle
                      defined by the `linestyle=` parameter.
        - linestyle: Matplotlib linestyle (e.g. '.', '-', '.-' etc.)
        - data_as_array: if sensors return arrays of values for different times
                         instead of values for a single time, put this
                         bool as True (default False).
                         NOTE: data_as array can also be a dict of bools
                         with names as keys if some sensors come as arrays
                         and some not.
        """
        self.names = names
        self.data_types = {name: data_types[name] for name in self.names}
        self.colors = colors
        self.legends = legends if legends is not None else {}
        self.linestyles = linestyles if linestyles is not None else {}
        self.linestyle = linestyle

        try:
            data_as_array.get   # no error if dict
        except AttributeError:  # it's a bool: put info for all sensors
            self.data_as_array = {name: data_as_array for name in self.names}
        else:
            self.data_as_array = data_as_array

        self.datalist_to_array = {}
        self.timelist_to_array = {}

        for name, data_as_array in self.data_as_array.items():
            if data_as_array:
                self.datalist_to_array[name] = self._list_of_value_arrays_to_array
                self.timelist_to_array[name] = self._list_of_time_arrays_to_array
            else:
                self.datalist_to_array[name] = self._list_of_single_values_to_array
                self.timelist_to_array[name] = self._list_of_single_times_to_array

        self.create_axes()
        self.set_colors()
        self.format_graph()
        self.fig.tight_layout()

        # Create onclick callback to activate / deactivate autoscaling
        self.cid = self.fig.canvas.mpl_connect('button_press_event',
                                               self.onclick)

    @property
    def all_data_types(self):
        """Return a set of all datatypes corresponding to the active names."""
        all_types = ()
        for name in self.names:
            data_types = self.data_types[name]
            all_types += data_types
        return set(all_types)

    @staticmethod
    def onclick(event):
        """Activate/deactivate autoscale by clicking to allow for data inspection.

        - Left click (e.g. when zooming, panning, etc.): deactivate autoscale
        - Right click: reactivate autoscale.
        """
        ax = event.inaxes
        if ax is None:
            pass
        elif event.button == 1:                        # left click
            ax.axes.autoscale(False, axis='both')
        elif event.button == 3:                        # right click
            ax.axes.autoscale(True, axis='both')
        else:
            pass

    def _list_of_single_values_to_array(self, datalist):
        """How to convert list of single values to a numpy array.

        This is to transform measurements stored in self.current_values
        into an array manageable by matplotlib for plotting.

        Can be subclassed to adapt to applications."""
        return datalist

    def _list_of_single_times_to_array(self, timelist):
        """How to convert list of single times to a numpy array.

        This is to transform measurements stored in self.current_values
        into an array manageable by matplotlib for plotting.

        Can be subclassed to adapt to applications."""
        return timelist

    def _list_of_value_arrays_to_array(self, datalist):
        """How to convert list of arrays of values to a numpy array.

        This is to transform measurements stored in self.current_values
        into an array manageable by matplotlib for plotting.

        Can be subclassed to adapt to applications."""
        return np.concatenate(datalist)

    def _list_of_time_arrays_to_array(self, timelist):
        """How to convert list array of times to a numpy array.

        This is to transform measurements stored in self.current_values
        into an array manageable by matplotlib for plotting.

        Can be subclassed to adapt to applications."""
        return np.concatenate(timelist)

    def set_colors(self):
        """"Define fig/ax colors if supplied"""
        if self.colors is None:
            self.colors = {}
        else:
            figcolor = self.colors.get('fig', 'white')
            self.fig.set_facecolor(figcolor)
            for ax in self.axs.values():
                axcolor = self.colors.get('ax', 'white')
                ax.set_facecolor(axcolor)
                ax.grid()

        missing_color_names = []
        n_missing_colors = 0
        for name, dtypes in self.data_types.items():
            try:
                self.colors[name]
            except (KeyError, TypeError):
                missing_color_names.append(name)
                n_missing_colors += len(dtypes)

        if not n_missing_colors:
            return

        m = cm.get_cmap('tab10', n_missing_colors)
        i = 0
        for name in missing_color_names:
            dtypes = self.data_types[name]
            colors = []
            for _ in dtypes:
                colors.append(m.colors[i])
                i += 1
            self.colors[name] = tuple(colors)

    def create_lines(self):
        """Create lines for each value of each sensor"""
        self.lines = {}
        self.lines_list = []

        for name in self.names:

            dtypes = self.data_types[name]
            clrs = self.colors[name]
            labels = self.legends.get(name, [None] * len(dtypes))
            lstyles = self.linestyles.get(name, [self.linestyle] * len(dtypes))

            self.lines[name] = []

            for dtype, clr, label, lstyle in zip(dtypes, clrs, labels, lstyles):

                # Plot data in correct axis depending on type
                ax = self.axs[dtype]
                line, = ax.plot([], [], lstyle, color=clr, label=label)

                self.lines[name].append(line)
                # Below, used for returning animated artists for blitting
                self.lines_list.append(line)

        if self.legends:
            for ax in self.axs.values():
                ax.legend(loc='lower left')

    def create_empty_data(self):
        data = {}
        for name in self.names:
            times = []
            values = []
            for _ in self.data_types[name]:
                values.append([])
            data[name] = {'times': times, 'values': values}
        return data

    def create_axes(self):
        """To be defined in subclasses. Returns fig, axs"""
        pass

    def format_graph(self):
        """To be defined in subclasses."""
        pass


class UpdateGraphBase:

    def __init__(self, graph, q_plot,
                 e_stop=None, e_close=None, e_graph=None,
                 dt_graph=1, blit=False):
        """Update plot with data received from a queue.

        INPUTS
        ------
        - graph: object of GraphBase class and subclasses
        - q_plot: dict {name: queue} with sensor names and data queues
        - e_stop (optional): external stop request, closes the figure if set
        - e_close (optional) is set when the figure has been closed
        - e_graph (optional) is set when the graph is activated
        - dt_graph: time interval to update the graph
        - blit: if True, use blitting to speed up the matplotlib animation

        Attention, if the figure is closed, the e_close event is triggered so
        do not put in e_close a threading event that is supposed to stay alive
        even if the figure gets closed. Rather, use the e_stop event.
        """
        self.graph = graph
        self.q_plot = q_plot
        self.e_stop = e_stop
        self.e_close = e_close
        self.e_graph = e_graph
        self.dt_graph = dt_graph
        self.blit = blit

        self.graph.fig.canvas.mpl_connect('close_event', self.on_fig_close)

    def on_fig_close(self, event):
        """When figure is closed, set threading events accordingly."""
        if self.e_close:
            self.e_close.set()
        if self.e_graph:
            self.e_graph.clear()

    def plot_new_data(self, i):
        """define what to do at each loop of the matplotlib animation."""

        if self.e_stop:
            if self.e_stop.is_set():
                plt.close(self.graph.fig)
                # since figure is closed, e_close and e_graph are taken care of
                # by the on_fig_close() function

        self.before_getting_measurements()

        for queue in self.q_plot.values():
            while True:
                try:
                    measurement = queue.get(timeout=0)
                except Empty:
                    break
                else:
                    self.manage_measurement(measurement)

        self.after_getting_measurements()

        if self.blit:
            return self.graph.animated_artists

    def _manage_data(self, data):
        """What to do with individual measurements coming from live data.

        (Measurement already formatted into data by format_measurement())
        Define in subclass
        """
        pass

    def manage_measurement(self, measurement):
        """Wrapper around _manage_data()."""
        # The line below allows some sensors to avoid being plotted by reading
        # None when called.
        if measurement is None:
            return

        data = self.graph.format_measurement(measurement)
        self._manage_data(data)

    def before_getting_measurements(self):
        """Anything to do before measurements from queue have been processed.

        Define in subclass (optional)
        """
        pass

    def after_getting_measurements(self):
        """Anything to do when all measurements from queue have been processed.

        (i.e. manage_measurement() has been called for all elements in queue.)
        Define in subclass (optional)
        """
        pass

    def run(self):

        # Below, it does not work if there is no value = before the FuncAnimation
        ani = FuncAnimation(fig=self.graph.fig,
                            func=self.plot_new_data,
                            interval=self.dt_graph * 1000,
                            cache_frame_data=False,
                            save_count=0,
                            blit=self.blit)

        plt.show(block=True)  # block=True allow the animation to work even
        # when matplotlib is in interactive mode (plt.ion()).

        return ani



"""Command line parser"""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>


import argparse


def parse_function(functions,
                   package_name=None,
                   description='',
                   default_function=None):
    """Trigger a function with arguments from command line.

    Parameters
    ----------
    - functions: dict {name: function} (name is used in the CLI to refer
                 to the function and can be the function full name or a
                 shortcut)
    - package name: optional, to print help correctly when using parser in
                    __main__.py file of a package.
                    (see e.g. https://bugs.python.org/issue22240)
    - description: any specific description of functions/arguments to add
                   to the default commmand line parser help
    - default_function: name of the function (functions dict key) that is
                        triggered if no function name is given (optional).
    """
    if package_name is not None:
        add_kwargs = {'prog': f"python -m {package_name}"}

    parser = argparse.ArgumentParser(description=description,
                                     formatter_class=argparse.RawTextHelpFormatter,
                                     **add_kwargs)

    # The nargs='?' is to have a positional argument with a default value
    msg_func = f"\nFunction to trigger.\nAvailable: {', '.join(functions)}."
    add_kwargs = {}
    if default_function is not None:
        msg_func += f"\nDefault: {default_function}."
        add_kwargs['default'] = default_function

    parser.add_argument('function', type=str, nargs='?', help=msg_func, **add_kwargs)

    msg_args = '\nArgs/kwargs of function.\nNo quotes required around str.'
    msg_args += "\nOther types inferred.\nExample foo a=hello for foo(a='hello')"
    parser.add_argument('arguments', type=str, nargs='*', help=msg_args)

    parsed = parser.parse_args()

    args = []
    kwargs = {}
    for argument in parsed.arguments:
        raw_arg = argument.split('=')
        try:
            k, v = raw_arg
        except ValueError:
            a, = raw_arg
            try:
                # Avoids having to explicily put type
                exec(f"args.append({a})")
            except (NameError, SyntaxError):
                # In case of strings, use the input as the string itself
                args.append(a)
        else:
            try:
                # Same strategy as above
                exec(f"kwargs['{k}'] = {v}")
            except (NameError, SyntaxError):
                kwargs[k] = v

    try:
        return functions[parsed.function](*args, **kwargs)
    except KeyError:
        # If default function not supplied, argparse thinks the first argument
        # is the function name
        try:
            args.append(parsed.function)
            return functions[default_function](*args, **kwargs)
        except Exception as e:
            print(f'Unknown command (Error: {e})')



"""Control classes for making programs/cycles of device/recording settings."""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>


# Standard library
from datetime import datetime
from threading import Thread, Event
from math import inf
from pathlib import Path

# Other modules
import oclock

# Local imports
from .program import _format_time, _get_and_check_input
from .program import Program, Stairs, Teeth

# ----------------------------------------------------------------------------
# ================================ Base Class ================================
# ----------------------------------------------------------------------------


class Control:
    """Base class to manage temporal evolution of settings.

    Can be used for ramping and programming temporal patterns of pressure,
    temperature, etc.

    Needs to be subclassed.
    """
    # [Optional] allowed inputs as parameters for control (e.g. 'p', 'rh', etc.)
    # To be defined in subclasses as an iterable, if necessary
    possible_inputs = None

    def __init__(self,
                 range_limits,
                 round_digits=3,
                 print_log=False,
                 save_log=True,
                 log_file='Control_Log.txt',
                 savepath='.'):
        """Init control object, with parameters:

        - range_limits: limits in settable parameter value (tuple min, max)
        - round_digits: number of digits after decimal point to keep/consider
                        when reading or applying settings
        - dt: time interval (s) between commands
        - range_limits: limits in settable parameter value (tuple min, max).
        - round_digits: number of digits after decimal point to keep/consider
                        when reading or applying settings
        - print_log: if True (default), print succession of settings in console
        - save_log: if True (default), save succession of settings sent to
                    device into a .txt file.
        - log_file: name of .txt file in which to save log of settings.
                    (default: Control_Log.txt)
        - savepath: directory in which to save the log file
        """
        self.range_limits = range_limits
        self.round_digits = round_digits
        self.print_log = print_log
        self.log_file = Path(savepath) / log_file if save_log else None

    # -------- Private methods for class operation behind the scenes ---------

    def _check_range_limits(self, qty, value, message=True):
        """Return value if within limits, else return higher or lower limit."""
        vmin, vmax = self.range_limits
        value_min = vmin if vmin is not None else -inf
        value_max = vmax if vmax is not None else inf

        if value_min <= value <= value_max:
            return value
        else:
            if value < value_min:
                value_setpt = value_min
            else:
                value_setpt = value_max
            if message:
                msg = (f'Required {qty}={value} outside of allowed '
                       f'range {value_min}-{value_max}. '
                       f'Setting kept at {value_setpt}.')
                self._manage_message(msg)
            return value_setpt

    def _print_ramp_info(self, qty, v1, v2, duration):
        """Print information about new program step in console."""
        msg = f'==> NEW STEP from {qty}={v1} to {qty}={v2} in {duration}'
        self._manage_message(msg)

    def _manage_message(self, msg, force_print=False):
        """Print in console and/or save to log file if options are activated"""

        t_str = datetime.now().isoformat(sep=' ', timespec='seconds')
        line = f'[{t_str}] {msg}\n'

        if self.log_file:
            try:
                with open(self.log_file, 'a', encoding='utf8') as file:
                    file.write(line)
            except Exception as e:
                print(f'Error saving to log file: {e}')

        if self.print_log or force_print:
            print(line)

    # ------- Private methods that need to be defined in child classes -------

    def _convert_input(self, **values):
        """Convert input from possible_inputs into parameter usable by device.

        For example, if one inputs p=... for a circulated thermal bath, convert
        the vapor pressure p to the temperature of water generating that bath,
        using the dewpoint at that pressure.

        To be defined in subclasses.
        """
        pass

    # ---------- Public methods that need to be defined in subclasses---------

    def ramp(self, duration, **values):
        """Ramp from val1 to val2 with given duration.

        =================== Must be defined in subclasses! ===================
        It is better if this method is non blocking, for user experience.

        Parameters
        ----------
        - duration: timedelta or 'h:m:s' str
        - values: dict (kwargs) with
            - key(word): setting quantity (e.g., p=, rh=, aw=, T=)
            - value: tuple (start, stop): start and end values of ramp.

        Example
        -------
        >>> Control().ramp(':1:30', rh=(50, 30))
        generates a ramp going from 50%RH to 30%RH (at 25°C) in 1.5 hours.
        """
        pass

    def stop(self):
        """Stop control (e.g. cancel ramp, stop timers, etc.).

        For example, cancel timer and cancel blocking behavior, etc.
        """
        pass

    # ------------- Factory methods to generate program objects --------------

    def program(self, repeat=1, **steps):
        """Convenience method to generate a program. see Program Class."""
        return Program(control=self,
                       repeat=repeat,
                       **steps)

    def stairs(self, duration=None, repeat=1, **steps):
        """Convenience method to generate a stairs program. see Stairs Class."""
        return Stairs(control=self,
                      duration=duration,
                      repeat=repeat,
                      **steps)

    def teeth(self, slope=None, slope_unit='/min', plateau_duration=None,
              start='plateau', repeat=1, **steps):
        """Convenience method to generate a teeth program. see Teeth Class."""
        return Teeth(control=self,
                     slope=slope,
                     slope_unit=slope_unit,
                     plateau_duration=plateau_duration,
                     start=start,
                     repeat=repeat,
                     **steps)


# ----------------------------------------------------------------------------
# ======== Base subclass for control using periodic update of settings ========
# ----------------------------------------------------------------------------


class PeriodicControl(Control):
    """Control using periodic update of device setting, no feedback"""

    def __init__(self,
                 dt=1,
                 range_limits=(None, None),
                 round_digits=3,
                 print_log=False,
                 save_log=True,
                 log_file='Control_Log.txt',
                 savepath='.'):
        """Create PeriodicControl object, with parameters:

        - dt: time interval (s) between commands
        - range_limits: limits in settable parameter value (tuple min, max).
        - round_digits: number of digits after decimal point to keep/consider
                        when reading or applying settings
        - print_log: if True (default), print succession of settings in console
        - save_log: if True (default), save succession of settings sent to
                    device into a .txt file
        - log_file: name of .txt file in which to save log of settings.
                    (default: Control_Log.txt)
        - savepath: directory in which to save the log file
        """
        Control.__init__(self,
                         range_limits=range_limits,
                         round_digits=round_digits,
                         print_log=print_log,
                         save_log=save_log,
                         log_file=log_file,
                         savepath=savepath)
        self._dt = dt
        self.timer = oclock.Timer(interval=dt)
        self.stop_event = Event()

    # ============================== Properties ==============================

    @property
    def dt(self):
        return self._dt

    @dt.setter
    def dt(self, value):
        self.timer.interval = value
        self._dt = value

    # ================== Methods to be defined in subclasses =================

    def _convert_input(self, **values):
        """Convert input from possible_inputs into parameter usable by device.

        (see Control class)
        """
        pass

    def _apply_setting(self, value):
        """"Define how to apply a setting to the device of interest.

        Defined in subclasses.
        e.g. device.setpt = value
        """
        pass

    def _read_setting(self):
        """"Define how to read the setting on the device of interest.

        (e.g. to check that setting has been really applied)
        Defined in subclasses.
        e.g. return device.setpt
        """
        pass

    def _print_setting(self, value):
        """How to print information about current setting in console.

        (OPTIONAL)
        Define in subclass.
        """
        return f'Setting: {value}'

    # =============== Methods deriving from the methods above ================

    def apply_setting(self, value):
        setting = round(value, self.round_digits)
        self._apply_setting(setting)

    def read_setting(self):
        setting = self._read_setting()
        return round(setting, self.round_digits)

    def print_setting(self, value):
        rounded_setting = round(value, self.round_digits)
        msg = self._print_setting(rounded_setting)
        self._manage_message(msg)

    # =========== MISC. methods used for ramping and set settings ============

    def _check_range_and_apply_setting(self, qty, value):
        """Check setting ok, rescale it if not, then apply to device."""
        target_setting = self._convert_input(**{qty: value})
        final_setting = self._check_range_limits(qty, target_setting)
        try:
            self.apply_setting(final_setting)
        except Exception:
            t_str = datetime.now().isoformat(sep=' ', timespec='seconds')
            print(f'Impossible to apply setting {qty}={value} ({t_str}).')
        else:
            self.print_setting(final_setting)

    def _try_read_setting(self):
        """Try to read setting from device."""
        try:
            setting = self.read_setting()
        except Exception:
            t_str = datetime.now().isoformat(sep=' ', timespec='seconds')
            print(f'Impossible to read setting ({t_str}).')
        else:
            return setting

    # ============================ Ramping methods ===========================

    def _apply_setting_and_check_done(self, qty, value, attempts=10):
        """Stay at a given value setting for the quantity of interest (blocking)

        Not a public method; is used by _ramp() and ramp().
        """
        setting = self._convert_input(**{qty: value})
        target_setting = self._check_range_limits(qty, setting, message=False)

        for _ in range(attempts):

            self._check_range_and_apply_setting(qty, value)
            actual_setting = self._try_read_setting()
            target_round = round(target_setting, self.round_digits)

            if actual_setting == target_round:
                return

            # This is to be able to stop the program even when the system
            # is continuously trying to apply a setting.
            if self.stop_event.is_set():
                return

            self.timer.checkpt()

        else:
            msg = 'WARNING -- Could not apply setting: '
            msg += f'target setting {target_round} and actual setting '
            msg += f'{actual_setting} do not match.'
            self._manage_message(msg, force_print=True)

    def _ramp(self, duration, **values):
        """Ramp from val1 to val2 with given duration.

        This method is BLOCKING. See ramp() for the nonblocking version and
        for parameter info and examples.
        """
        self.stop_event.clear()

        qty, (v1, v2) = _get_and_check_input(
            values,
            possible_inputs=self.possible_inputs
        )
        self._print_ramp_info(qty, v1, v2, duration)

        self.timer.reset()
        t_ramp = _format_time(duration)

        if v1 == v2:
            # If dwelling, no need to update the setting regularly.
            # Just apply it once, but make sure it has really been applied.
            dwell = True
            self._apply_setting_and_check_done(qty, v2)
            self._manage_message(f'Dwelling started ({qty}={v2})')
        else:
            dwell = False

        while self.timer.elapsed_time <= t_ramp:

            if not dwell:
                t = self.timer.elapsed_time
                setting = v1 + t / t_ramp * (v2 - v1) if t_ramp > 0 else v2

                self._check_range_and_apply_setting(qty, setting)

            self.timer.checkpt()

            if self.stop_event.is_set():
                self._manage_message('==X Manual STOP')
                return

        else:
            if dwell:
                self._manage_message('Dwelling finished')
            if not dwell:
                self._check_range_and_apply_setting(qty, v2)
                # below, avoids taking two datapoints in a row for programs
                self.timer.checkpt()

    def ramp(self, duration, **values):
        """Ramp from val1 to val2 with given duration.

        =================== Must be defined in subclasses! ===================
        It is better if this method is non blocking, for user experience.

        Parameters
        ----------
        - duration: timedelta or 'h:m:s' str
        - values: dict (kwargs) with
            - key(word): setting quantity (e.g., p=, rh=, aw=, T=)
            - value: tuple (start, stop): start and end values of ramp.

        Example
        -------
        >>> Control().ramp(':1:30', rh=(50, 30))
        generates a ramp going from 50%RH to 30%RH (at 25°C) in 1.5 hours.
        """
        Thread(target=self._ramp, args=(duration,), kwargs=values).start()

    def stop(self):
        """Cancel timers and stop ramp."""
        self.stop_event.set()
        self.timer.stop()


# ----------------------------------------------------------------------------
# =========== Periodic control adapted to the prevo.record module ============
# ----------------------------------------------------------------------------


class RecordingControl(PeriodicControl):
    """Control of prevo Recordings objects"""

    def __init__(self,
                 recording=None,
                 ppty=None,
                 dt=1,
                 range_limits=(None, None),
                 round_digits=3,
                 print_log=False,
                 save_log=True,
                 log_file='Control_Log_Recording.txt',
                 savepath='.'):
        """Create PeriodicTemperatureControl object, with parameters:

        - recording: Recording object or subclass.
        - ppty: ControlledProperty object describing the property to control
                within that recording
        - dt: time interval (s) between commands
        - range_limits: safety limits; tuple (min, max)
        - round_digits: number of digits after decimal point to keep/consider
                        when reading or applying settings
        - print_log: if True (default), print succession of settings in console
        - save_log: if True (default), save succession of settings sent to
                    device into a .txt file
        - log_file: name of .txt file in which to save log of settings.
                    (default: Control_Log_Recording.txt)
        - savepath: directory in which to save the log file
        """
        self.ppty = ppty
        self.possible_inputs = self.ppty.commands

        PeriodicControl.__init__(self,
                                 dt=dt,
                                 range_limits=range_limits,
                                 round_digits=round_digits,
                                 print_log=print_log,
                                 save_log=save_log,
                                 log_file=log_file,
                                 savepath=savepath)

        self.recording = recording

    def _apply_setting(self, value):
        """Set property value on recording."""
        exec(f'self.recording.{self.ppty.attribute} = {value}')

    def _read_setting(self):
        """Get property value of recording."""
        self._value = None  # exec does not work on local scope
        exec(f'self._value = self.recording.{self.ppty.attribute}')
        return self._value

    def _print_setting(self, value):
        """How to print information about current setting in console."""
        return f'Setting: {self.ppty.attribute}={value}'

    def _convert_input(self, **values):
        for command in self.ppty.commands:
            try:
                return values[command]
            except KeyError:
                pass
        else:
            input_key, = values.keys()
            raise ValueError(f'Input {input_key} does not match {self.ppty.commands}.')



"""Plot data from sensors in oscilloscope-like fashion"""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>


# Standard library imports
import time

# Non standard imports
import numpy as np
import matplotlib.pyplot as plt

# Local imports
from .general import NumericalGraphBase, UpdateGraphBase


class UpdateGraph(UpdateGraphBase):

    def _manage_data(self, data):

        tmin, tmax = self.graph.get_time_boundaries(data)

        if self.graph.reference_time is None:
            self.graph.reference_time = tmin   # Take time of 1st data as time 0

        self.graph.update_stored_data(data=data,
                                      stored_data=self.graph.current_data)

        # In case measurement arrives late after window has already refreshed,
        # duplicate it to previous data so that it is remains visible
        # This is particularly useful for data arriving as arrays; in which
        # case the array will be duplicated to appear both at the beginning
        # and end of the window when the times in the arrays span values
        # across the window wrapping time.
        if tmin < self.graph.reference_time:
            self.graph.update_stored_data(data=data,
                                          stored_data=self.graph.previous_data)

        # There is no need to do the same for 'future' points that would arrive
        # with tmax > reference_time + window_size, because in
        # principle all data arriving is from the past or present.

    def after_getting_measurements(self):
        self.graph.update_lines()
        self.graph.update_bars()
        if self.graph.reference_time and (self.graph.relative_time > self.graph.window_width):
            self.graph.wrap()


class OscilloGraph(NumericalGraphBase):

    def __init__(self,
                 names,
                 data_types,
                 data_ranges,
                 window_width=10,
                 colors=None,
                 legends=None,
                 linestyles=None,
                 linestyle='.',
                 data_as_array=False):
        """Initiate figures and axes for data plot as a function of asked types.

        Input
        -----
        - names: iterable of names of recordings/sensors that will be plotted.
        - data_types: dict with the recording names as keys, and the
                      corresponding data types as values.
                      (dict can have more keys than those in 'names')
        - data_ranges: dict with the possible data types as keys, and the
                       corresponding range of values expected for this data
                       as values. Used to set ylims of graph initially.
                      (dict can have more keys than actual data types used)
        - window_width: width (in seconds) of the displayed window
        - colors: optional dict of colors with keys 'fig', 'ax', 'bar' and the
                  names of the recordings.
        - legends: optional dict of legend names (iterable) corresponding to
                   all channels of each sensor, with the names of the
                   recordings as keys.
        - linestyles: optional dict of linestyles (iterable) to distinguish
                      channels and sensors, with the names of the recordings
                      as keys. If not specified (None), all lines have the
                      linestyle defined by the `linestyle=` parameter (see
                      below). If only some recordings are specified, the other
                      recordings have the default linestyle or the linestyle
                      defined by the `linestyle=` parameter.
        - linestyle: Matplotlib linestyle (e.g. '.', '-', '.-' etc.)
        - data_as_array: if sensors return arrays of values for different times
                         instead of values for a single time, put this
                         bool as True (default False)
                         NOTE: data_as array can also be a dict of bools
                         with names as keys if some sensors come as arrays
                         and some not.
        """
        self.data_ranges = data_ranges
        self.window_width = window_width
        self.reference_time = None

        super().__init__(names=names,
                         data_types=data_types,
                         colors=colors,
                         legends=legends,
                         linestyles=linestyles,
                         linestyle=linestyle,
                         data_as_array=data_as_array)

        self.current_data = self.create_empty_data()
        self.previous_data = self.create_empty_data()

    # ========================== Misc. Init Methods ==========================

    def create_axes(self):
        """Generate figure/axes as a function of input data types"""

        n = len(self.all_data_types)
        self.fig, axes = plt.subplots(n, 1)

        # Transform axes into a tuple if only one ax
        try:
            iter(axes)
        except TypeError:
            axes = axes,

        self.axs = {}
        for ax, datatype in zip(axes, self.all_data_types):
            ax.set_ylabel(datatype)
            self.axs[datatype] = ax

    def format_graph(self):
        """Set colors, time formatting, etc."""

        w = self.window_width
        for dtype, ax in self.axs.items():
            ax.set_xlim((-0.001 * w, 1.001 * w))
            ax.set_ylim(self.data_ranges[dtype])
            ax.grid()

        self.create_lines()
        self.create_bars()

    def create_bars(self):
        """Create traveling bars"""
        self.bars = {}
        for dtype, ax in self.axs.items():
            barcolor = self.colors.get('bar', 'silver')
            bar = ax.axvline(0, linestyle='-', c=barcolor, linewidth=4)
            self.bars[dtype] = bar

    # =============== Methods overriden from the parent class ================

    def _list_of_single_values_to_array(self, datalist):
        """To be able to filter on time condition and concatenate"""
        return np.array(datalist, dtype=np.float64)

    def _list_of_single_times_to_array(self, timelist):
        """To be able to filter on time condition and concatenate"""
        return np.array(timelist, dtype=np.float64)

    def on_click():
        """Here turn off autoscale, which can cause problems with blitting."""
        pass

    # ======= Methods that can be subclassed to adapt to applications ========

    def format_measurement(self, measurement):
        """Transform measurement from the queue into something usable by manage_data()

        Can be subclassed to adapt to various applications.
        Here, assumes data is incoming in the form of a dictionary with at
        least keys:
        - 'name' (str, identifier of sensor)
        - 'time (unix)' (float or array of floats)
        - 'values' (iterable of values, or iterable of arrays of values)

        Subclass to adapt to applications.
        """
        return measurement

    def get_time_boundaries(self, data):
        """Subclass if necessary."""
        t = data['time (unix)']
        name = data['name']

        if self.data_as_array[name]:
            return t[0], t[-1]
        else:
            return t, t

    # ========================== Misc. properties ============================

    @property
    def current_time(self):
        return time.time()

    @property
    def relative_time(self):
        return self.current_time - self.reference_time

    @property
    def animated_artists(self):
        artists = self.lines_list + list(self.bars.values())
        return artists

    # ========================= Graph Update Methods =========================

    def wrap(self):
        """What to do each time the bars exceeed window size"""
        self.previous_data = self.current_data
        self.current_data = self.create_empty_data()
        self.reference_time += self.window_width

    def update_stored_data(self, data, stored_data):
        """Store measurement time and values in active data lists.

        Parameters
        ----------
        data: data as output by format_measurement()
        stored_data: either self.current_data or self.previous_data.
        """
        name = data['name']
        time = data['time (unix)']
        values = data['values']

        stored_data[name]['times'].append(time)
        for i, value in enumerate(values):
            stored_data[name]['values'][i].append(value)

    def update_lines(self):
        """Update line positions with current data."""

        # Keep only previous drawings after current bar position

        for name in self.lines:

            lines = self.lines[name]
            previous_data = self.previous_data[name]
            current_data = self.current_data[name]

            rel_times = []

            # Avoids problems if no data stored yet
            prev_exists = bool(previous_data['times'])
            curr_exists = bool(current_data['times'])

            if not (prev_exists or curr_exists):
                continue

            if curr_exists:
                curr_times = self.timelist_to_array[name](current_data['times'])
                curr_rel_times = curr_times - self.reference_time
                rel_times.append(curr_rel_times)

            if prev_exists:
                prev_times = self.timelist_to_array[name](previous_data['times'])
                prev_condition = (prev_times + self.window_width > self.current_time)
                prev_rel_times = prev_times[prev_condition] - self.reference_time + self.window_width
                rel_times.append(prev_rel_times)

            rel_times_array = np.concatenate(rel_times)

            for line, prev_values, curr_values in zip(lines,
                                                      previous_data['values'],
                                                      current_data['values']):

                vals = []

                if curr_exists:
                    curr_vals = self.datalist_to_array[name](curr_values)
                    vals.append(curr_vals)

                if prev_exists:
                    prev_vals = self.datalist_to_array[name](prev_values)
                    vals.append(prev_vals[prev_condition])

                values_array = np.concatenate(vals)

                line.set_data(rel_times_array, values_array)

    def update_bars(self):
        if self.reference_time:   # Avoids problems if no data arrived yet
            t = self.relative_time
            for bar in self.bars.values():
                bar.set_xdata(t)

    def run(self, q_plot, e_stop=None, e_close=None, e_graph=None,
            dt_graph=0.02, blit=True):
        """Run live view of oscilloscope with data from queues.

        (Convenience method to instantiate a UpdateGraph object)

        Parameters
        ----------
        - q_plot: dict {name: queue} with sensor names and data queues
        - e_stop (optional): external stop request, closes the figure if set
        - e_close (optional) is set when the figure has been closed
        - e_graph (optional) is set when the graph is activated
        - dt graph: time interval to update the graph
        - blit: if True, use blitting to speed up the matplotlib animation
        """
        update_oscillo = UpdateGraph(graph=self,
                                     q_plot=q_plot,
                                     e_stop=e_stop,
                                     e_close=e_close,
                                     e_graph=e_graph,
                                     dt_graph=dt_graph,
                                     blit=blit)
        update_oscillo.run()



"""Record several sensors as a function of time with interactive CLI and graph."""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>


# Non-standard imports
import gittools
import prevo

# Local imports
from .general import RecordBase, RecordingBase
from ..csv import CsvFile
from ..plot import NumericalGraph


class NumericalRecording(RecordingBase):
    """Recording class that saves numerical sensor data to csv files."""

    def __init__(self,
                 Sensor,
                 filename,
                 path='.',
                 column_names=None,
                 column_formats=None,
                 csv_separator='\t',
                 **kwargs):
        """Init NumericalRecording object.

        Parameters
        ----------
        - Sensor: subclass of SensorBase.
        - filename: name of csv/tsv file in which data will be saved.
        - column_names: iterable of str (name of columns in csv file)
        - column_formats: iterable of str (optional, str formatting)
        - csv_separator: character to separate columns in CSV file.

        Additional kwargs from RecordingBase:
        - path: directory in which csv/tsv file will be created.
        - dt: time interval between readings (default 1s).
        - ctrl_ppties: optional iterable of properties (ControlledProperty
                       objects) to control on the recording in addition to
                       default ones (time interval and active on/off)
        - active: if False, do not record data until self.active set to True.
        - continuous: if True, take data as fast as possible from sensor.
        - warnings: if True, print warnings of Timer (e.g. loop too short).
        - precise: if True, use precise timer in oclock (see oclock.Timer).
        - programs: iterable of programs, which are object of the
                    prevo.control.Program class or subclasses.
                    --> optional pre-defined temporal pattern of change of
                    properties of recording (e.g. define some times during
                    which sensor is active or not, or change time interval
                    between data points after some time, etc.)
        - control_params: dict {command: kwargs} containing any kwargs to pass
                          to the program controls (e.g. dt, range_limits, etc.)
                          Note: if None, use default params
                          (see RecordingControl class)
        """
        super().__init__(Sensor=Sensor, path=path, **kwargs)

        self.file_manager = CsvFile(filename=filename,
                                    column_names=column_names,
                                    column_formats=column_formats,
                                    path=path,
                                    csv_separator=csv_separator)


    def format_measurement(self, measurement):
        """Format raw sensor data into a measurement object.

        This object is put in the saving queue and in the plotting queue
        if active.

        Here we assume that measurements are in the form of dictionaries, but
        this can be changed by subclassing.
        """

        if measurement is None:
            return
        # Adding the name is useful e.g. for plotting.
        measurement['name'] = self.name
        return measurement

    def save(self, measurement, file):
        """Save to file

        Inputs
        ------
        - measurement: Measurement object
        - file:

        Output
        ------
        Iterable of data to be saved in CSV file

        The length of the iterable must be equal to that of column_names.

        Here, we assume a standard measurement as a dict with keys
        'time (unix)', 'dt (s)', 'values'

        Can be redefined in subclasses.
        NOTE: if measurement is None, Record.data_save() does not save the data
        """
        time_info = (measurement['time (unix)'], measurement['dt (s)'])
        value_info = measurement['values']
        self.file_manager._write_line(time_info + value_info, file=file)


class NumericalRecord(RecordBase):
    """Class managing simultaneous temporal recordings of numerical sensors"""

    def __init__(self,
                 recordings,
                 metadata_filename='Numerical_Metadata.json',
                 checked_modules=(),
                 data_types=None,
                 dt_graph=0.1,
                 graph_colors=None,
                 dirty_ok=True,
                 **kwargs):
        """Init NumericalRecord object.

        Parameters
        ----------
        - recordings: iterable of recording objects (RecordingBase or subclass)
        - metadata_filename: name of .json file in which metadata is saved
        - checked_modules: iterable of python modules, the version of which
                           will be saved in metadata, in addition to prevo.
        - data_types: iterable of data types (for selecting window in which
                      which to plot data when graphs are active);
                      if not supplied, graphs can not be instantiated.
        - dt_graph: time interval to refresh numerical graph.
        - graph_colors: dict of graph colors for numerical graph
                        (see prevo.plot)
        - dirty_ok: if False, record cannot be started if git repositories are
                    not clean (commited).

        Additional kwargs from RecordingBase:
        - path: directory in which data is recorded.
        - on_start: optional iterable of objects with a .start() or .run()
                    method, that need to be started at the same time as
                    Record.start().
                    Note: for now, start() and run() need to be non-blocking.
        - dt_save: how often (in seconds) queues are checked and written to files
                   (it is also how often files are open/closed)
        - dt_request: time interval (in seconds) for checking user requests
                      (e.g. graph pop-up)
        - dt_check: time interval (in seconds) for checking queue sizes.
        """
        super().__init__(recordings=recordings, **kwargs)
        self.metadata_filename = metadata_filename
        self.checked_modules = set((prevo,) + tuple(checked_modules))

        # Graphing options -------------
        self.data_types = data_types
        self.dt_graph = dt_graph
        self.graph_colors = graph_colors

        self.dirty_ok = dirty_ok
        self.get_numerical_recordings()

    def get_numerical_recordings(self):
        """Useful when vacuum combined with other recording types (e.g. camrec)"""
        self.numerical_recordings = {}
        for name, recording in self.recordings.items():
            if issubclass(recording.__class__, NumericalRecording):
                self.numerical_recordings[name] = recording

    def _save_metadata(self, filename):
        """To call save_metadata() with custom filenames"""
        metadata_file = self.path / filename

        if metadata_file.exists():
            metadata_file = self.increment_filename(metadata_file)

        gittools.save_metadata(metadata_file,
                               module=self.checked_modules,
                               dirty_warning=True,
                               dirty_ok=self.dirty_ok,
                               notag_warning=True,
                               nogit_ok=True,
                               nogit_warning=True)

    def save_metadata(self):
        """Save info on code version of modules used for the recording"""
        self._save_metadata(filename=self.metadata_filename)

    def data_plot(self):
        """What to do when graph event is triggered"""
        if self.data_types is None:
            print('WARNING --- No data types supplied. Graph not possible.')
            self.e_graph.clear()
            return

        graph = NumericalGraph(names=self.numerical_recordings,
                               data_types=self.data_types,
                               colors=self.graph_colors,
                               data_as_array=False)

        # In case the queue contains other measurements than numerical
        # (e.g. images from cameras)
        numerical_queue = {name: self.q_plot[name]
                           for name in self.numerical_recordings}

        graph.run(q_plot=numerical_queue,
                  e_stop=self.e_stop,
                  e_graph=self.e_graph,
                  dt_graph=self.dt_graph)



"""Plot data from sensors (from live measurements or saved data)."""

# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>


# Standard library imports
from datetime import datetime
from threading import Event, Thread
from queue import Queue
from pathlib import Path

# Non standard imports
import tzlocal
import numpy as np

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.lines import Line2D
import oclock

from .general import NumericalGraphBase, UpdateGraphBase
from ..record import SensorError

# The two lines below have been added following a console FutureWarning:
# "Using an implicitly registered datetime converter for a matplotlib plotting
# method. The converter was registered by pandas on import. Future versions of
# pandas will require you to explicitly register matplotlib converters."
try:
    import pandas as pd
    from pandas.plotting import register_matplotlib_converters
    register_matplotlib_converters()
except ModuleNotFoundError:
    pandas_available = False
else:
    pandas_available = True


# Misc =======================================================================


local_timezone = tzlocal.get_localzone()


# =============================== Main classes ===============================


class UpdateGraph(UpdateGraphBase):

    def _manage_data(self, data):
        self.graph.update_current_data(data)

    def after_getting_measurements(self):
        self.graph.update_lines()
        self.graph.update_time_formatting()


class NumericalGraph(NumericalGraphBase):

    def __init__(self,
                 names,
                 data_types,
                 colors=None,
                 legends=None,
                 linestyles=None,
                 linestyle='.',
                 data_as_array=False,
                 time_conversion='numpy'):
        """Initiate figures and axes for data plot as a function of asked types.

        Input
        -----
        - names: iterable of names of recordings/sensors that will be plotted.
        - data types: dict with the recording names as keys, and the
                      corresponding data types as values.
                      (dict can have more keys than those in 'names')
        - colors: optional dict of colors with keys 'fig', 'ax', and the
                  names of the recordings.
        - legends: optional dict of legend names (iterable) corresponding to
                   all channels of each sensor, with the names of the
                   recordings as keys.
        - linestyles: optional dict of linestyles (iterable) to distinguish
                      channels and sensors, with the names of the recordings
                      as keys. If not specified (None), all lines have the
                      linestyle defined by the `linestyle=` parameter (see
                      below). If only some recordings are specified, the other
                      recordings have the default linestyle or the linestyle
                      defined by the `linestyle=` parameter.
        - linestyle: Matplotlib linestyle (e.g. '.', '-', '.-' etc.)
        - data_as_array: if sensors return arrays of values for different times
                         instead of values for a single time, put this
                         bool as True (default False)
        - time_conversion: how to convert from unix time to datetime for arrays;
                           possible values: 'numpy', 'pandas'.
        """
        self.timezone = local_timezone

        super().__init__(names=names,
                         data_types=data_types,
                         colors=colors,
                         legends=legends,
                         linestyles=linestyles,
                         linestyle=linestyle,
                         data_as_array=data_as_array)

        time_converters = {'datetime': self._to_datetime_datetime,
                           'numpy': self._to_datetime_numpy,
                           'pandas': self._to_datetime_pandas}

        self.time_converters = {}
        for name in self.names:
            if self.data_as_array[name]:
                self.time_converters[name] = time_converters[time_conversion]
            else:
                self.time_converters[name] = time_converters['datetime']

        self.current_data = self.create_empty_data()  # For live updates

    def create_axes(self):
        """Generate figure/axes as a function of input data types"""

        if len(self.all_data_types) == 4:
            fig, axes = plt.subplots(2, 2, figsize=(10, 9))
        if len(self.all_data_types) == 3:
            fig, axes = plt.subplots(1, 3, figsize=(14, 4))
        elif len(self.all_data_types) == 2:
            fig, axes = plt.subplots(1, 2, figsize=(10, 4))
        elif len(self.all_data_types) == 1:
            fig, ax = plt.subplots(1, 1, figsize=(6, 4))
            axes = ax,
        else:
            msg = f'Mode combination {self.all_data_types} not supported yet'
            raise Exception(msg)

        axs = {}
        for ax, datatype in zip(axes, self.all_data_types):
            ax.set_ylabel(datatype)
            axs[datatype] = ax

        self.fig = fig
        self.axs = axs

    def format_graph(self):
        """Set colors, time formatting, etc."""

        # Concise formatting of time -----------------------------------------

        self.locator = {}
        self.formatter = {}

        for ax in self.axs.values():
            self.locator[ax] = mdates.AutoDateLocator(tz=self.timezone)
            self.formatter[ax] = mdates.ConciseDateFormatter(self.locator,
                                                             tz=self.timezone)

        # Create line objects for every sensor channel -----------------------
        self.create_lines()

    @property
    def animated_artists(self):
        return self.lines_list

    # ============================= Main methods =============================

    def _to_datetime_datetime(self, unix_time):
        """Transform single value of unix time into timezone-aware datetime."""
        return datetime.fromtimestamp(unix_time, self.timezone)

    @staticmethod
    def _to_datetime_numpy(unix_times):
        """Transform iterable / array of unix times into datetimes.

        Note: this is the fastest method, but the datetimes are in UTC format
              (not local time)
        """
        return (np.array(unix_times) * 1e9).astype('datetime64[ns]')

    def _to_datetime_pandas(self, unix_times):
        """Transform iterable / array of datetimes into pandas Series.

        Note: here, the datetimes are in local timezone format, but this is
              slower than the numpy approach.
        """
        # For some reason, it's faster (and more precise) to convert to numpy first
        np_times = (np.array(unix_times) * 1e9).astype('datetime64[ns]')
        pd_times = pd.Series(np_times)
        return pd.to_datetime(pd_times, utc=True).dt.tz_convert(self.timezone)

    def format_measurement(self, measurement):
        """Transform measurement from the queue into something usable by manage_data()

        Can be subclassed to adapt to various applications.
        Here, assumes data is incoming in the form of a dictionary with at
        least keys:
        - 'name' (str, identifier of sensor)
        - 'time (unix)' (floar or array of floats)
        - 'values' (iterable of values, or iterable of arrays of values)

        Subclass to adapt to applications.
        """
        data = {key: measurement[key] for key in ('name', 'values')}

        name = measurement['name']
        t_unix = measurement['time (unix)']

        data['time'] = self.time_converters[name](t_unix)

        return data

    # For static plots -------------------------------------------------------

    def _plot(self, data):
        """Generic plot method for static data.

        data is a dict obtained from format_measurement(measurement)
        where measurement is an object from the data queue.
        """
        name = data['name']
        values = data['values']
        time = data['time']

        dtypes = self.data_types[name]  # all data types for this specific signal
        clrs = self.colors[name]

        for value, dtype, clr in zip(values, dtypes, clrs):
            ax = self.axs[dtype]  # plot data in correct axis depending on type
            ax.plot(time, value, self.linestyle, color=clr)

        # Use Concise Date Formatting for minimal space used on screen by time
        different_types = set(dtypes)
        for dtype in different_types:
            ax = self.axs[dtype]
            ax.xaxis.set_major_locator(self.locator[ax])
            ax.xaxis.set_major_formatter(self.formatter[ax])

    # For updated plots ------------------------------------------------------

    def update_current_data(self, data):
        """Store measurement time and values in active data lists."""

        name = data['name']
        values = data['values']
        time = data['time']

        self.current_data[name]['times'].append(time)
        for i, value in enumerate(values):
            self.current_data[name]['values'][i].append(value)

    def update_lines(self):
        """Update line positions with current data."""

        for name in self.names:

            lines = self.lines[name]
            current_data = self.current_data[name]

            if current_data['times']:  # Avoids problems if no data stored yet

                times = self.timelist_to_array[name](current_data['times'])

                for line, curr_values in zip(lines,
                                             current_data['values']):

                    values = self.datalist_to_array[name](curr_values)
                    line.set_data(times, values)

    def update_time_formatting(self):
        """Use Concise Date Formatting for minimal space used on screen by time"""
        for ax in self.axs.values():
            ax.xaxis.set_major_locator(self.locator[ax])
            ax.xaxis.set_major_formatter(self.formatter[ax])
            # Lines below are needed for autoscaling to work
            ax.relim()
            ax.autoscale_view(tight=True, scalex=True, scaley=True)

    def run(self, q_plot, e_stop=None, e_close=None, e_graph=None,
            dt_graph=0.1, blit=False):
        """Run live view of plot with data from queues.

        (Convenience method to instantiate a UpdateGraph object)

        Parameters
        ----------
        - q_plot: dict {name: queue} with sensor names and data queues
        - e_stop (optional): external stop request, closes the figure if set
        - e_close (optional) is set when the figure has been closed
        - e_graph (optional) is set when the graph is activated
        - dt graph: time interval to update the graph
        - blit: if True, use blitting to speed up the matplotlib animation
        """
        update_plot = UpdateGraph(graph=self,
                                  q_plot=q_plot,
                                  e_stop=e_stop,
                                  e_close=e_close,
                                  e_graph=e_graph,
                                  dt_graph=dt_graph,
                                  blit=blit)
        update_plot.run()


# ============== Classes using Graph-like objects to plot data ===============


# ------------------------------- Static graph -------------------------------


class PlotSavedData:
    """Class to create graphs from saved data."""

    def __init__(self, names, graph, SavedData, file_names, path='.'):
        """Parameters:

        - names: names of sensors/recordings to consider
        - graph: object of GraphBase class and subclasses
        - SavedData: subclass of measurement.SavedDataBase
                            must have (name, filename, path) as arguments
                            and must define load() and format_as_measurement()
                            (see measurements.py)
        - file_names: dict {name: filename (str)} of files containing data
        - path: directory in which data is saved
        """
        self.names = names
        self.graph = graph
        self.SavedData = SavedData
        self.file_names = file_names
        self.path = Path(path)

    def show(self):
        """Static plot of saved data"""
        for name in self.names:
            saved_data = self.SavedData(name,
                                        filename=self.file_names[name],
                                        path=self.path)
            saved_data.load()
            measurement = saved_data.format_as_measurement()
            self.graph.plot(measurement)
        self.graph.fig.tight_layout()
        plt.show(block=False)


# ------------------------------ Updated graphs ------------------------------


class PlotUpdatedData:
    """Class to initiate and manage periodic reading of queues to plot data.

    Is subclassed to define get_data(), which indicates how to get the data
    (e.g. from live sensors, or from reading a file, etc.)
    """

    def __init__(self, graph, dt_data=1, dt_graph=1):
        """Parameters:

        - graph: object of GraphBase class and subclasses
        - dt_data is how often (in s) the loop checks for (or gets) new data.
        - dt_graph is how often (in s) the graph animation is updated
        """
        self.graph = graph
        self.timer = oclock.Timer(interval=dt_data, name='Data Update')
        self.dt_graph = dt_graph

        self.queues = {name: Queue() for name in self.names}
        self.e_graph = Event()
        self.e_stop = Event()

    # Methods that need to be defined in subclasses --------------------------

    def get_data(self, name):
        """Get data and put it in queue"""
        pass

    # Other methods ----------------------------------------------------------

    def run(self):
        """Run reading of all data sources concurrently"""

        for name in self.names:
            Thread(target=self.get_data, args=(name,)).start()

        self.e_graph.set()  # This is supposed to be set when graph is active.

        # e_stop two times, because we want a figure closure event to also
        # trigger stopping of the recording process here.

        self.graph.run(q_plot=self.queues,
                       e_stop=self.e_stop,
                       e_close=self.e_stop,
                       e_graph=self.e_graph,
                       dt_graph=self.dt_graph)


class PlotLiveSensors(PlotUpdatedData):
    """Create live graph by reading the sensors directly."""

    def __init__(self, graph, recordings, dt_data=1, dt_graph=1):
        """Parameters:

        - graph: object of GraphBase class and subclasses
        - recordings: dict {name: recording(RecordingBase) object}
        - dt_data: how often (in s) sensors are probed
        - dt_graph is how often (in s) the graph animation is updated
        """
        self.recordings = recordings
        self.names = list(recordings)

        super().__init__(graph=graph, dt_data=dt_data, dt_graph=dt_graph)

    def get_data(self, name):
        """Check if new data is read by sensor, and put it in data queue."""
        self.timer.reset()
        recording = self.recordings[name]

        with recording.Sensor() as sensor:

            while not self.e_stop.is_set():
                try:
                    data = sensor.read()
                except SensorError:
                    pass
                else:
                    measurement = recording.format_measurement(data)
                    recording.after_measurement()
                    self.queues[name].put(measurement)
                    self.timer.checkpt()


class PlotSavedDataUpdated(PlotUpdatedData, PlotSavedData):
    """Extends PlotSavedData to be able to periodically read file to update."""

    def __init__(self, names, graph, SavedData, file_names,
                 path='.', dt_data=1, dt_graph=1):
        """Parameters:

        - names: names of sensors/recordings to consider
        - graph: object of GraphBase class and subclasses
        - SavedData: Measurement class that manages data loading
                     must have (name, filename, path) as arguments and
                     must define load(), format_as_measurement() and
                     number_of_measurements() methods (see measurements.py)
        - file_names: dict {name: filename (str)} of files containing data
        - path: directory in which data is saved
        - dt_data: how often (in s) the files are checked for updates.
        - dt_graph: how often (in s) the graph animation is updated.
        """
        PlotSavedData.__init__(self,
                               names=names,
                               graph=graph,
                               SavedData=SavedData,
                               file_names=file_names,
                               path=path)

        PlotUpdatedData.__init__(self,
                                 graph=graph,
                                 dt_data=dt_data,
                                 dt_graph=dt_graph)

    def get_data(self, name):
        """Check if new data is added to file, and put it in data queue."""
        self.timer.reset()

        saved_data = self.SavedData(name,
                                    filename=self.file_names[name],
                                    path=self.path)

        n0 = saved_data.number_of_measurements()

        while not self.e_stop.is_set():

            n = saved_data.number_of_measurements()

            if n > n0:
                saved_data.load(nrange=(n0 + 1, n))
                if saved_data.data is not None:
                    measurement = saved_data.format_as_measurement()
                    self.queues[name].put(measurement)
                    n0 = n

            self.timer.checkpt()



# ----------------------------- License information --------------------------

# This file is part of the prevo python package.
# Copyright (C) 2022 Olivier Vincent

# The prevo package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# The prevo package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with the prevo python package.
# If not, see <https://www.gnu.org/licenses/>


from setuptools import setup

if __name__ == '__main__':
    setup(use_scm_version=True)



"""Tests for the prevo package.

(VERY partial)
"""

# Standard library
from pathlib import Path

# Non standard
import pytest

# local imports
import prevo
from prevo.measurements import SavedCsvData


datafolder = Path(prevo.__file__).parent / '..' / 'data/manip'

t_column, dt_column = 'time (unix)', 'dt (s)'

names = 'P', 'T', 'B1'
filenames = {'P': 'Vacuum_Pressure.tsv',
             'T': 'Vacuum_Temperature.tsv',
             'B1': 'Vacuum_SourceBath.tsv'}

meas_numbers = {'P': 22553, 'T': 11271, 'B1': 9883}

n = 8  # measurement line number to check when loading full file
# (considering first data line is numbered 1)

nrange = (5, 10)  # range for partial loading test
nred = n - nrange[0] + 1  # corresponding line in the partial data

lines = {'P': (1616490450.506, 0.167, 2727.25),
         'T': (1616490515.961, 2.623, 26.8932, 25.4829),
         'B1': (1616490514.585, 0.091, 24.2640)}

measP = SavedCsvData('P', filenames['P'], datafolder)
measT = SavedCsvData('T', filenames['T'], path=datafolder)
measB1 = SavedCsvData(name='B1', filename=filenames['B1'],
                      path=datafolder, csv_separator='\t')

saved_datas = {'P': measP, 'T': measT, 'B1': measB1}


def test_number_of_measurements():  # test with a variety of call options
    ns = [sdata.number_of_measurements() for sdata in saved_datas.values()]
    assert ns == list(meas_numbers.values())


@pytest.mark.parametrize('name', names)
def test_load_full(name):             # test full loading of data from file
    sdata = saved_datas[name]
    sdata.load()
    assert len(sdata.data) == meas_numbers[name]
    assert tuple(sdata.data.loc[n - 1].round(decimals=4)) == lines[name]


@pytest.mark.parametrize('name', names)
def test_load_partial(name):
    sdata = saved_datas[name]
    sdata.load(nrange)  # test partial loading of data
    assert len(sdata.data) == nrange[1] - nrange[0] + 1
    assert tuple(sdata.data.loc[nred - 1].round(decimals=4)) == lines[name]
